import inspect
import math
import os.path as op
from contextlib import nullcontext as does_not_raise

import pytest
import torch

import plenoptic as po
from conftest import DEVICE, check_loss_saved_synth


def rgb_mse(*args):
    return po.metric.mse(*args).mean()


def rgb_l2_norm(*args):
    return po.tools.optim.l2_norm(*args).mean()


# MAD requires metrics are *dis*-similarity metrics, so that they
# return 0 if two images are identical (SSIM normally returns 1)
def dis_ssim(*args):
    return (1 - po.metric.ssim(*args)).mean()


class ModuleMetric(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.mdl = po.simul.Gaussian((31, 31)).to(DEVICE)

    def forward(self, x, y):
        return (self.mdl(x) - self.mdl(y)).abs().mean()


class NonModuleMetric:
    def __init__(self):
        self.name = "nonmodule"

    def __call__(self, x, y):
        return (x - y).abs().sum()


class TestMAD:
    @pytest.mark.parametrize("target", ["min", "max"])
    @pytest.mark.parametrize("model_order", ["mse-ssim", "ssim-mse"])
    @pytest.mark.parametrize("store_progress", [False, True, 2])
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_basic(self, curie_img, target, model_order, store_progress):
        if model_order == "mse-ssim":
            model = po.metric.mse
            model2 = dis_ssim
        elif model_order == "ssim-mse":
            model = dis_ssim
            model2 = po.metric.mse
        mad = po.synth.MADCompetition(
            curie_img, model, model2, target, metric_tradeoff_lambda=1
        )
        mad.synthesize(max_iter=5, store_progress=store_progress)
        if store_progress:
            mad.synthesize(max_iter=5, store_progress=store_progress)

    @pytest.mark.parametrize(
        "fail",
        [
            False,
            "img",
            "metric1",
            "metric2",
            "target",
            "tradeoff",
            "range_penalty",
            "allowed_range",
        ],
    )
    @pytest.mark.parametrize("range_penalty", [0.1, 0])
    @pytest.mark.parametrize("allowed_range", [(0, 1), (-1, 1)])
    @pytest.mark.parametrize("rgb", [False, True])
    @pytest.mark.filterwarnings(
        "ignore:SSIM was designed for grayscale images:UserWarning"
    )
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_save_load(
        self, curie_img, fail, range_penalty, allowed_range, rgb, tmp_path
    ):
        # this works with either rgb or grayscale images
        metric = rgb_mse
        if rgb:
            curie_img = curie_img.repeat(1, 3, 1, 1)
            metric2 = rgb_l2_norm
        else:
            metric2 = dis_ssim
        target = "min"
        tradeoff = 1
        mad = po.synth.MADCompetition(
            curie_img,
            metric,
            metric2,
            target,
            metric_tradeoff_lambda=tradeoff,
            allowed_range=allowed_range,
            range_penalty_lambda=range_penalty,
        )
        mad.synthesize(max_iter=4, store_progress=True)
        mad.save(op.join(tmp_path, "test_mad_save_load.pt"))
        if fail:
            if fail == "img":
                curie_img = torch.rand_like(curie_img)
                expectation = pytest.raises(
                    ValueError,
                    match="Saved and initialized attribute image have different values",
                )
            elif fail == "metric1":
                # this works with either rgb or grayscale images (though note
                # that SSIM just operates on each RGB channel independently,
                # which is probably not the right thing to do)
                metric = dis_ssim
                expectation = pytest.raises(
                    ValueError,
                    match=(
                        "Saved and initialized optimized_metric output have different"
                        " values"
                    ),
                )
            elif fail == "metric2":
                # this works with either rgb or grayscale images
                metric2 = rgb_mse
                expectation = pytest.raises(
                    ValueError,
                    match=(
                        "Saved and initialized reference_metric output have different"
                        " values"
                    ),
                )
            elif fail == "target":
                target = "max"
                expectation = pytest.raises(
                    ValueError,
                    match="Saved and initialized minmax are different",
                )
            elif fail == "tradeoff":
                tradeoff = 10
                expectation = pytest.raises(
                    ValueError,
                    match=(
                        "Saved and initialized metric_tradeoff_lambda are different"
                    ),
                )
            elif fail == "allowed_range":
                allowed_range = (0, 5)
                expectation = pytest.raises(
                    ValueError,
                    match=("Saved and initialized allowed_range are different"),
                )
            elif fail == "range_penalty":
                range_penalty = 0.5
                expectation = pytest.raises(
                    ValueError,
                    match=("Saved and initialized range_penalty_lambda are different"),
                )
            mad_copy = po.synth.MADCompetition(
                curie_img,
                metric,
                metric2,
                target,
                metric_tradeoff_lambda=tradeoff,
                allowed_range=allowed_range,
                range_penalty_lambda=range_penalty,
            )
            with expectation:
                mad_copy.load(
                    op.join(tmp_path, "test_mad_save_load.pt"),
                    map_location=DEVICE,
                )
        else:
            mad_copy = po.synth.MADCompetition(
                curie_img,
                metric,
                metric2,
                target,
                metric_tradeoff_lambda=tradeoff,
                allowed_range=allowed_range,
                range_penalty_lambda=range_penalty,
            )
            mad_copy.load(
                op.join(tmp_path, "test_mad_save_load.pt"), map_location=DEVICE
            )
            # check that can resume
            mad_copy.synthesize(max_iter=5, store_progress=True)
        if rgb:
            # since this is a fixture, get this back to a grayscale image
            curie_img = curie_img.mean(1, True)

    def test_setup_initial_noise(self, einstein_img):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.setup(0.5)
        mad.synthesize(5)

    def test_setup_fail(self, einstein_img):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.setup()
        with pytest.raises(ValueError, match=r"setup\(\) can only be called once"):
            mad.setup()

    def test_setup_load_fail(self, einstein_img, tmp_path):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=4)
        mad.save(op.join(tmp_path, "test_mad_setup_load_fail.pt"))
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.load(op.join(tmp_path, "test_mad_setup_load_fail.pt"))
        with pytest.raises(
            ValueError, match="Cannot set initial_noise after calling load"
        ):
            mad.setup(0.5)

    @pytest.mark.filterwarnings("ignore:You will need to call setup:UserWarning")
    def test_synth_then_setup(self, einstein_img, tmp_path):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.setup(optimizer=torch.optim.SGD)
        mad.synthesize(max_iter=4)
        mad.save(op.join(tmp_path, "test_mad_synth_then_setup.pt"))
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.load(op.join(tmp_path, "test_mad_synth_then_setup.pt"))
        with pytest.raises(ValueError, match="Don't know how to initialize"):
            mad.synthesize(5)
        mad.setup(optimizer=torch.optim.SGD)
        mad.synthesize(5)

    def test_load_init_fail(self, einstein_img, tmp_path):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=4, store_progress=True)
        mad.save(op.join(tmp_path, "test_mad_load_init_fail.pt"))
        with pytest.raises(
            ValueError, match="load can only be called with a just-initialized"
        ):
            mad.load(op.join(tmp_path, "test_mad_load_init_fail.pt"))

    @pytest.mark.parametrize("fail", [False, "name", "behavior"])
    def test_load_names(self, fail, einstein_img, tmp_path):
        # name and behavior same, but module path is different
        if fail is False:

            def mse(x, y):
                return po.tools.optim.mse(x, y)

            metric2 = mse
            expectation = does_not_raise()
        # name different but behavior same
        elif fail == "name":

            def bad_metric(x, y):
                return po.tools.optim.mse(x, y)

            metric2 = bad_metric
            expectation = pytest.raises(
                ValueError,
                match="Saved and initialized optimized_metric have different names",
            )
        # name same but behavior different
        elif fail == "behavior":

            def mse(x, y):
                return po.tools.optim.l2_norm(x, y)

            metric2 = mse
            expectation = pytest.raises(
                ValueError,
                match=(
                    "Saved and initialized optimized_metric output have different"
                    " values"
                ),
            )
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=4, store_progress=True)
        mad.save(op.join(tmp_path, f"test_mad_load_names_{fail}.pt"))
        mad = po.synth.MADCompetition(
            einstein_img,
            metric2,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        with expectation:
            mad.load(op.join(tmp_path, f"test_mad_load_names_{fail}.pt"))

    def test_examine_saved_object(self, einstein_img, tmp_path):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=4, store_progress=True)
        mad.save(op.join(tmp_path, "test_mad_examine.pt"))
        po.tools.examine_saved_synthesis(op.join(tmp_path, "test_mad_examine.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("synth_type", ["eig", "met"])
    def test_load_object_type(self, einstein_img, model, synth_type, tmp_path):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=4, store_progress=True)
        mad.save(op.join(tmp_path, "test_mad_load_object_type.pt"))
        if synth_type == "eig":
            mad = po.synth.Eigendistortion(einstein_img, model)
        elif synth_type == "met":
            mad = po.synth.Metamer(einstein_img, model)
        with pytest.raises(
            ValueError, match="Saved object was a.* but initialized object is"
        ):
            mad.load(op.join(tmp_path, "test_mad_load_object_type.pt"))

    @pytest.mark.parametrize("metric_behav", ["dtype", "shape", "name"])
    @pytest.mark.parametrize("metric", ["optimized", "reference"])
    def test_load_metric_change(self, einstein_img, metric, metric_behav, tmp_path):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=4, store_progress=True)
        mad.save(op.join(tmp_path, "test_mad_load_metric_change.pt"))

        def new_metric(x, y):
            if metric_behav == "dtype":
                return po.tools.optim.mse(x, y).to(torch.float64)
            elif metric_behav == "shape":
                return torch.stack(
                    [po.tools.optim.mse(x, y) for _ in range(2)]
                ).unsqueeze(0)
            elif metric_behav == "name":
                if metric == "optimized":
                    return po.tools.optim.mse(x, y)
                elif metric == "reference":
                    return po.tools.optim.l2_norm(x, y)

        if metric_behav == "name":
            expectation_str = (
                f"Saved and initialized {metric}_metric have different names"
            )
        elif metric_behav == "shape":
            # this gets raised during the metric validation step
            expectation_str = "metric should return a scalar value but output had shape"
        else:
            expectation_str = (
                f"Saved and initialized {metric}_metric output have different"
                f" {metric_behav}"
            )
        with pytest.raises(ValueError, match=expectation_str):
            if metric == "optimized":
                mad = po.synth.MADCompetition(
                    einstein_img,
                    new_metric,
                    po.tools.optim.l2_norm,
                    "min",
                    metric_tradeoff_lambda=1,
                )
            elif metric == "reference":
                mad = po.synth.MADCompetition(
                    einstein_img,
                    po.metric.mse,
                    new_metric,
                    "min",
                    metric_tradeoff_lambda=1,
                )
            mad.load(op.join(tmp_path, "test_mad_load_metric_change.pt"))

    @pytest.mark.parametrize(
        "model", ["frontend.LinearNonlinear.nograd"], indirect=True
    )
    @pytest.mark.parametrize("attribute", ["saved", "init"])
    def test_load_attributes(self, einstein_img, model, attribute, tmp_path):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=4, store_progress=True)
        if attribute == "saved":
            mad.test = "BAD"
            err_str = "Saved"
        mad.save(op.join(tmp_path, "test_mad_load_attributes.pt"))
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        if attribute == "init":
            mad.test = "BAD"
            err_str = "Initialized"
        with pytest.raises(
            ValueError, match=rf"{err_str} object has 1 attribute\(s\) not present"
        ):
            mad.load(op.join(tmp_path, "test_mad_load_attributes.pt"))

    @pytest.mark.parametrize(
        "optim_opts",
        [
            None,
            "SGD",
            "SGD-args",
            "Adam",
            "Adam-args",
            "Scheduler",
            "Scheduler-args",
            "LBFGS",
            "LBFGS-args",
        ],
    )
    @pytest.mark.parametrize("fail", [True, False])
    @pytest.mark.filterwarnings("ignore:You will need to call setup:UserWarning")
    def test_load_optimizer(self, curie_img, optim_opts, fail, tmp_path):
        mad = po.synth.MADCompetition(
            curie_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        scheduler = None
        optimizer = None
        optimizer_kwargs = None
        scheduler_kwargs = None
        check_optimizer = [torch.optim.Adam, {"eps": 1e-8, "lr": 0.01}]
        check_scheduler = None
        if optim_opts is not None:
            if "Scheduler" in optim_opts:
                scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
                check_scheduler = [
                    torch.optim.lr_scheduler.ReduceLROnPlateau,
                    {"factor": 0.1},
                ]
                if "args" in optim_opts:
                    scheduler_kwargs = {"factor": 1e-3}
                    check_scheduler[1] = {"factor": 1e-3}
            else:
                if "Adam" in optim_opts:
                    optimizer = torch.optim.Adam
                elif "SGD" in optim_opts:
                    optimizer = torch.optim.SGD
                    check_optimizer[0] = torch.optim.SGD
                    check_optimizer[1] = {"lr": 0.01}
                elif "LBFGS" in optim_opts:
                    optimizer = torch.optim.LBFGS
                    check_optimizer[0] = torch.optim.LBFGS
                    check_optimizer[1] = {"lr": 0.01}
                if "args" in optim_opts:
                    optimizer_kwargs = {"lr": 1}
                    check_optimizer[1] = {"lr": 1}
        mad.setup(
            optimizer=optimizer,
            scheduler=scheduler,
            optimizer_kwargs=optimizer_kwargs,
            scheduler_kwargs=scheduler_kwargs,
        )
        mad.synthesize(max_iter=5)
        mad.save(op.join(tmp_path, "test_mad_optimizer.pt"))
        mad = po.synth.MADCompetition(
            curie_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.load(op.join(tmp_path, "test_mad_optimizer.pt"))
        optimizer_kwargs = None
        scheduler_kwargs = None
        if not fail:
            if optim_opts is not None:
                if "Adam" in optim_opts:
                    optimizer = torch.optim.Adam
                elif "SGD" in optim_opts:
                    optimizer = torch.optim.SGD
                elif "LBFGS" in optim_opts:
                    optimizer = torch.optim.LBFGS
                if "Scheduler" in optim_opts:
                    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
            expectation = does_not_raise()
        else:
            expect_str = "User-specified optimizer must have same type"
            if optim_opts is None:
                optimizer = torch.optim.SGD
            else:
                if optim_opts == "Adam":
                    optimizer = torch.optim.SGD
                elif optim_opts == "Adam-args":
                    optimizer = torch.optim.Adam
                    optimizer_kwargs = {"lr": 1}
                    expect_str = (
                        "When initializing optimizer after load, optimizer_kwargs"
                    )
                elif optim_opts == "SGD":
                    optimizer = None
                    expect_str = "Don't know how to initialize saved optimizer"
                elif optim_opts == "SGD-args":
                    optimizer = torch.optim.SGD
                    optimizer_kwargs = {"lr": 1}
                    expect_str = (
                        "When initializing optimizer after load, optimizer_kwargs"
                    )
                elif optim_opts == "LBFGS":
                    optimizer = None
                    expect_str = "Don't know how to initialize saved optimizer"
                elif optim_opts == "LBFGS-args":
                    optimizer = torch.optim.LBFGS
                    optimizer_kwargs = {"lr": 1}
                    expect_str = (
                        "When initializing optimizer after load, optimizer_kwargs"
                    )
                elif optim_opts == "Scheduler":
                    scheduler = torch.optim.lr_scheduler.ConstantLR
                    expect_str = "User-specified scheduler must have same type"
                elif optim_opts == "Scheduler-args":
                    scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
                    scheduler_kwargs = {"factor": 1e-3}
                    expect_str = (
                        "When initializing scheduler after load, scheduler_kwargs"
                    )
            expectation = pytest.raises(ValueError, match=expect_str)
        # these fail during setup
        with expectation:
            mad.setup(
                optimizer=optimizer,
                scheduler=scheduler,
                optimizer_kwargs=optimizer_kwargs,
                scheduler_kwargs=scheduler_kwargs,
            )
            mad.synthesize(max_iter=5)
            if not isinstance(mad.optimizer, check_optimizer[0]):
                raise ValueError("Didn't properly set optimizer!")
            state_dict = mad.optimizer.state_dict()["param_groups"][0]
            for k, v in check_optimizer[1].items():
                if state_dict[k] != v:
                    raise ValueError(
                        "Didn't properly set optimizer kwargs! "
                        f"Expected {v} but got {state_dict[k]}!"
                    )
            if check_scheduler is not None:
                if not isinstance(mad.scheduler, check_scheduler[0]):
                    raise ValueError("Didn't properly set scheduler!")
                state_dict = mad.scheduler.state_dict()
                for k, v in check_scheduler[1].items():
                    if mad.scheduler.state_dict()[k] != v:
                        raise ValueError("Didn't properly set scheduler kwargs!")
            elif mad.scheduler is not None:
                raise ValueError("Didn't set scheduler to None!")

    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_load_tol(self, einstein_img, tmp_path):
        mad = po.synth.MADCompetition(
            einstein_img,
            po.metric.mse,
            lambda *args: 1 - po.metric.ssim(*args),
            "min",
            metric_tradeoff_lambda=1,
            allowed_range=(-1, 2),
        )
        mad.synthesize(5)
        mad.save(op.join(tmp_path, "test_mad_load_tol.pt"))
        mad = po.synth.MADCompetition(
            einstein_img + 1e-7 * torch.rand_like(einstein_img),
            po.metric.mse,
            lambda *args: 1 - po.metric.ssim(*args),
            "min",
            metric_tradeoff_lambda=1,
            allowed_range=(-1, 2),
        )
        with pytest.raises(ValueError, match="Saved and initialized attribute image"):
            mad.load(op.join(tmp_path, "test_mad_load_tol.pt"))
        mad.load(op.join(tmp_path, "test_mad_load_tol.pt"), tensor_equality_atol=1e-7)

    @pytest.mark.parametrize(
        "optimizer",
        [
            "SGD",
            "SGD-args",
            "Adam",
            "Adam-args",
            None,
            "Scheduler-args",
            "Scheduler",
            "LBFGS",
            "LBFGS-args",
        ],
    )
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_optimizer(self, curie_img, optimizer):
        mad = po.synth.MADCompetition(
            curie_img,
            po.metric.mse,
            lambda *args: 1 - po.metric.ssim(*args),
            "min",
            metric_tradeoff_lambda=1,
        )
        optimizer = None
        scheduler = None
        optimizer_kwargs = None
        scheduler_kwargs = None
        check_optimizer = [torch.optim.Adam, {"eps": 1e-8, "lr": 0.01}]
        check_scheduler = None
        if optimizer == "Adam":
            optimizer = torch.optim.Adam
        elif optimizer == "Adam-args":
            optimizer = torch.optim.Adam
            optimizer_kwargs = {"eps": 1e-5}
            check_optimizer[1] = {"eps": 1e-5}
        elif optimizer == "SGD":
            optimizer = torch.optim.SGD
            check_optimizer = [torch.optim.SGD, {"lr": 0.01}]
        elif optimizer == "SGD-args":
            optimizer = torch.optim.SGD
            optimizer_kwargs = {"lr": 1}
            check_optimizer = [torch.optim.SGD, {"lr": 1}]
        elif optimizer == "LBFGS":
            optimizer = torch.optim.LBFGS
            check_optimizer = [torch.optim.LBFGS, {"lr": 0.01}]
        elif optimizer == "LBFGS-args":
            optimizer = torch.optim.LBFGS
            optimizer_kwargs = {"lr": 1, "history_size": 10}
            check_optimizer = [torch.optim.LBFGS, {"lr": 1, "history_size": 10}]
        elif optimizer == "Scheduler":
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
            check_scheduler = [
                torch.optim.lr_scheduler.ReduceLROnPlateau,
                {"factor": 0.1},
            ]
        elif optimizer == "Scheduler-args":
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau
            scheduler_kwargs = {"factor": 1e-3}
            check_scheduler = [
                torch.optim.lr_scheduler.ReduceLROnPlateau,
                {"factor": 1e-3},
            ]
        mad.setup(
            optimizer=optimizer,
            scheduler=scheduler,
            optimizer_kwargs=optimizer_kwargs,
            scheduler_kwargs=scheduler_kwargs,
        )
        mad.synthesize(max_iter=5)
        if not isinstance(mad.optimizer, check_optimizer[0]):
            raise ValueError("Didn't properly set optimizer!")
        state_dict = mad.optimizer.state_dict()["param_groups"][0]
        for k, v in check_optimizer[1].items():
            if state_dict[k] != v:
                raise ValueError(
                    "Didn't properly set optimizer kwargs! "
                    f"Expected {v} but got {state_dict[k]}!"
                )
        if check_scheduler is not None:
            if not isinstance(mad.scheduler, check_scheduler[0]):
                raise ValueError("Didn't properly set scheduler!")
            state_dict = mad.scheduler.state_dict()
            for k, v in check_scheduler[1].items():
                if mad.scheduler.state_dict()[k] != v:
                    raise ValueError("Didn't properly set scheduler kwargs!")
        elif mad.scheduler is not None:
            raise ValueError("Didn't set scheduler to None!")

    @pytest.mark.parametrize("metric", [po.metric.mse, ModuleMetric, NonModuleMetric])
    @pytest.mark.parametrize("to_type", ["dtype", "device"])
    def test_to(self, curie_img, metric, to_type):
        # if metric is not the po.metric.mse function above, initialize it here,
        # otherwise we can get a weird state-dependence
        if not inspect.isfunction(metric):
            metric = metric()
        mad = po.synth.MADCompetition(
            curie_img, metric, po.tools.optim.l2_norm, "min", metric_tradeoff_lambda=1
        )
        mad.synthesize(max_iter=5)
        if to_type == "dtype":
            mad.to(torch.float64)
            assert mad.initial_image.dtype == torch.float64
            assert mad.image.dtype == torch.float64
            assert mad.mad_image.dtype == torch.float64
        # can only run this one if we're on a device with CPU and GPU.
        elif to_type == "device" and DEVICE.type != "cpu":
            mad.to("cpu")
        # initial_image doesn't get used anywhere after init, so check it like
        # this
        mad.initial_image - mad.image
        mad.mad_image - mad.image
        mad.synthesize(max_iter=5)

    @pytest.mark.skipif(DEVICE.type == "cpu", reason="Only makes sense to test on cuda")
    def test_map_location(self, curie_img, tmp_path):
        mad = po.synth.MADCompetition(
            curie_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=4, store_progress=True)
        mad.save(op.join(tmp_path, "test_mad_map_location.pt"))
        mad_copy = po.synth.MADCompetition(
            curie_img.to("cpu"),
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        assert mad_copy.image.device.type == "cpu"
        mad_copy.load(op.join(tmp_path, "test_mad_map_location.pt"), map_location="cpu")
        assert mad_copy.mad_image.device.type == "cpu"
        mad_copy.synthesize(max_iter=4, store_progress=True)

    @pytest.mark.skipif(DEVICE.type == "cpu", reason="Only makes sense to test on cuda")
    def test_to_midsynth(self, curie_img):
        mad = po.synth.MADCompetition(
            curie_img,
            po.metric.mse,
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=4, store_progress=2)
        assert mad.image.device.type == "cuda"
        assert mad.mad_image.device.type == "cuda"
        mad.to("cpu")
        mad.synthesize(max_iter=4, store_progress=2)
        assert mad.image.device.type == "cpu"
        assert mad.mad_image.device.type == "cpu"
        mad.to("cuda")
        mad.synthesize(max_iter=4, store_progress=2)
        assert mad.image.device.type == "cuda"
        assert mad.mad_image.device.type == "cuda"

    # MAD can accept multiple images on the batch dimension, but the metrics
    # must return a single number. This means, effectively, that we can do
    # synthesis for e.g., video metrics, but that we cannot synthesize several
    # images in parallel
    def test_batch_synthesis(self, curie_img, einstein_img):
        img = torch.cat([curie_img, einstein_img], dim=0)
        mad = po.synth.MADCompetition(
            img,
            lambda *args: po.metric.mse(*args).mean(),
            po.tools.optim.l2_norm,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad.synthesize(max_iter=10)
        assert mad.mad_image.shape == img.shape, (
            "MAD image should have the same shape as input!"
        )

    @pytest.mark.parametrize("input_dim", [2, 3, 4, 5])
    @pytest.mark.filterwarnings("ignore:.*mostly been tested on 4d inputs:UserWarning")
    def test_dimensionality(self, einstein_img, input_dim):
        img = einstein_img.squeeze()
        while img.ndimension() < input_dim:
            img = img.unsqueeze(0)
        mad = po.synth.MADCompetition(img, rgb_mse, rgb_l2_norm, "min", 1)
        mad.synthesize(5)

    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    @pytest.mark.parametrize("store_progress", [True, 2, 3])
    def test_store_rep(self, einstein_img, store_progress):
        mad = po.synth.MADCompetition(
            einstein_img, po.metric.mse, dis_ssim, "min", metric_tradeoff_lambda=1
        )
        max_iter = 3
        if store_progress == 3:
            max_iter = 6
        mad.synthesize(max_iter=max_iter, store_progress=store_progress)
        check_loss_saved_synth(
            mad.losses,
            mad.saved_mad_image,
            max_iter,
            mad.objective_function,
            store_progress,
        )
        mad.synthesize(max_iter=max_iter, store_progress=store_progress)
        check_loss_saved_synth(
            mad.losses,
            mad.saved_mad_image,
            2 * max_iter,
            mad.objective_function,
            store_progress,
        )

    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_save_mad_empty(self, einstein_img):
        mad = po.synth.MADCompetition(
            einstein_img, po.metric.mse, dis_ssim, "min", metric_tradeoff_lambda=1
        )
        torch.equal(mad.saved_mad_image, torch.empty(0))
        mad.synthesize(max_iter=3)
        torch.equal(mad.saved_mad_image, mad.mad_image.to("cpu"))

    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_mad_empty_loss(self, einstein_img):
        mad = po.synth.MADCompetition(
            einstein_img, po.metric.mse, dis_ssim, "min", metric_tradeoff_lambda=1
        )
        assert mad.objective_function().numel() == 0
        torch.equal(mad.losses, torch.empty(0))
        mad.setup()
        assert isinstance(mad.objective_function(), torch.Tensor)
        assert mad.losses.numel() > 0
        mad.synthesize(max_iter=2)
        assert isinstance(mad.objective_function(), torch.Tensor)
        assert mad.losses.numel() > 0

    @pytest.mark.parametrize("iteration", [None, 0, -2, -3, 2, 1, 6, -7])
    @pytest.mark.parametrize("store_progress", [True, False, 2])
    @pytest.mark.parametrize("iteration_selection", ["floor", "ceiling", "round"])
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_mad_get_progress(
        self, einstein_img, iteration, store_progress, iteration_selection
    ):
        mad = po.synth.MADCompetition(
            einstein_img, po.metric.mse, dis_ssim, "min", metric_tradeoff_lambda=1
        )
        mad.synthesize(max_iter=5, store_progress=store_progress)
        if iteration_selection == "floor":
            func = math.floor
        elif iteration_selection == "ceiling":
            func = math.ceil
        else:
            func = round
        expected_dict = {}
        if iteration is None:
            expected_dict["losses"] = mad.losses[-1]
            expected_dict.update(
                {
                    "iteration": 5,
                    "gradient_norm": None,
                    "pixel_change_norm": None,
                    "reference_metric_loss": mad.reference_metric_loss[-1],
                    "optimized_metric_loss": mad.optimized_metric_loss[-1],
                }
            )
            if store_progress:
                expected_dict.update(
                    {
                        "saved_mad_image": mad.mad_image.to("cpu"),
                        "store_progress_iteration": 5,
                    }
                )
        elif iteration not in [6, -7]:
            expected_dict["losses"] = mad.losses[iteration]
            if iteration < 0:
                # add one to account for loss and these attributes being off by one
                expected_dict.update(
                    {
                        "iteration": 6 + iteration,
                        "gradient_norm": mad.gradient_norm[iteration + 1],
                        "pixel_change_norm": mad.pixel_change_norm[iteration + 1],
                        "reference_metric_loss": mad.reference_metric_loss[iteration],
                        "optimized_metric_loss": mad.optimized_metric_loss[iteration],
                    }
                )
                if store_progress:
                    iter = func((6 + iteration) / store_progress)
                    expected_dict.update(
                        {
                            "saved_mad_image": mad.saved_mad_image[iter],
                            "store_progress_iteration": iter * store_progress,
                        }
                    )
            else:
                expected_dict.update(
                    {
                        "iteration": iteration,
                        "gradient_norm": mad.gradient_norm[iteration],
                        "pixel_change_norm": mad.pixel_change_norm[iteration],
                        "reference_metric_loss": mad.reference_metric_loss[iteration],
                        "optimized_metric_loss": mad.optimized_metric_loss[iteration],
                    }
                )
                if store_progress:
                    iter = func(iteration / store_progress)
                    expected_dict.update(
                        {
                            "saved_mad_image": mad.saved_mad_image[iter],
                            "store_progress_iteration": iter * store_progress,
                        }
                    )
        if iteration in [6, -7]:
            expectation = pytest.raises(IndexError, match=".*out of bounds with.*")
        elif store_progress == 2 and iteration in [1, -3]:
            expectation = pytest.warns(
                UserWarning, match="loss iteration and iteration"
            )
        else:
            expectation = does_not_raise()
        with expectation:
            progress = mad.get_progress(iteration, iteration_selection)
            assert progress.keys() == expected_dict.keys()
            for k, v in progress.items():
                if isinstance(v, torch.Tensor):
                    assert torch.equal(v, expected_dict[k]), f"{k} not as expected!"
                else:
                    assert v == expected_dict[k], f"{k} not as expected!"

    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_continue(self, einstein_img):
        mad = po.synth.MADCompetition(
            einstein_img, po.metric.mse, dis_ssim, "min", metric_tradeoff_lambda=1
        )
        mad.synthesize(max_iter=3, store_progress=True)
        mad.synthesize(max_iter=3, store_progress=True)

    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_nan_loss(self, einstein_img):
        # clone to prevent NaN from showing up in other tests
        img = einstein_img.clone()
        mad = po.synth.MADCompetition(
            img, po.metric.mse, dis_ssim, "min", metric_tradeoff_lambda=1
        )
        mad.synthesize(max_iter=5)
        mad.image[..., 0, 0] = torch.nan
        with pytest.raises(ValueError, match="Found a NaN in loss during optimization"):
            mad.synthesize(max_iter=1)

    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_change_precision_save_load(self, einstein_img, tmp_path):
        # Identity model doesn't change when you call .to() with a dtype
        # (unlike those models that have weights) so we use it here
        mad = po.synth.MADCompetition(
            einstein_img, po.metric.mse, dis_ssim, "min", metric_tradeoff_lambda=1
        )
        mad.synthesize(max_iter=5)
        mad.to(torch.float64)
        assert mad.mad_image.dtype == torch.float64, "dtype incorrect!"
        mad.save(op.join(tmp_path, "test_change_prec_save_load.pt"))
        mad_copy = po.synth.MADCompetition(
            einstein_img.to(torch.float64),
            po.metric.mse,
            dis_ssim,
            "min",
            metric_tradeoff_lambda=1,
        )
        mad_copy.load(op.join(tmp_path, "test_change_prec_save_load.pt"))
        mad_copy.synthesize(max_iter=5)
        assert mad_copy.mad_image.dtype == torch.float64, "dtype incorrect!"

    @pytest.mark.filterwarnings("ignore:Loss has converged:UserWarning")
    @pytest.mark.filterwarnings("ignore:Image range falls outside:UserWarning")
    def test_stop_criterion(self, einstein_img):
        # checking that this hits the criterion and stops early, so set seed
        # for reproducibility
        po.tools.set_seed(0)
        mad = po.synth.MADCompetition(
            einstein_img, po.metric.mse, dis_ssim, "min", metric_tradeoff_lambda=0.1
        )
        # losses[-1] corresponds to the *current* loss (of mad.mad_image), not the loss
        # from the most recent iteration. so losses[-2] is the loss of the last
        # synthesis iteration.
        mad.synthesize(max_iter=15, stop_criterion=1e-3, stop_iters_to_check=5)
        assert abs(mad.losses[-6] - mad.losses[-2]) < 1e-3, (
            "Didn't stop when hit criterion!"
        )
        assert abs(mad.losses[-7] - mad.losses[-3]) > 1e-3, (
            "Stopped after hit criterion!"
        )



"""
Synthesis methods.

Classes to perform stimulus synthesis.
"""

# ignore unused imports
# ruff: noqa: F401
from .eigendistortion import Eigendistortion
from .mad_competition import MADCompetition
from .metamer import Metamer, MetamerCTF



import pathlib
import re
import sys

paths = []
for p in sys.argv[1:]:
    p = pathlib.Path(p)
    if p.is_dir():
        p = list(p.glob("**/*.md"))
    elif p.suffix == ".md":
        p = [p]
    else:
        p = []
    paths.extend(p)


fails = []
for p in paths:
    with open(p) as f:
        md = f.read()
    if not md.startswith("---\njupytext"):
        # then this isn't a markdown notebook
        continue
    filename = p.stem
    if not re.findall(f"{{nb-download}}`{filename}.ipynb`", md):
        fails.append(p)


if fails:
    print("The following markdown notebooks are missing a download notebook button!")
    for p in fails:
        print(p)
    sys.exit(1)



"""Various helpful utilities for plotting or displaying information."""
# numpydoc ignore=ES01

import warnings
from typing import Any, Literal

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pyrtools as pt
import torch

from .data import to_numpy


def _find_zoom(
    image_heights: list[int], image_widths: list[int], ax: mpl.axes.Axes
) -> float:
    """
    Find best-fitting zoom based on image and axes sizes.

    If images are bigger than ``ax``, then we figure out the largest float of form
    ``1/d``, where ``d`` is an integer. If ``ax`` is bigger than images, figure out the
    largest integer we can use.

    Parameters
    ----------
    image_heights, image_widths
        The last two dimensions of all images to plot.
    ax
        The existing axis we will use for imshow.

    Returns
    -------
    zoom
        Our best guess at zoom.
    """

    def find_zoom_helper(x: float, limit: float) -> float:
        """
        Find zoom that works. This is only for limit < x.

        Parameters
        ----------
        x
            The sizes to consider.
        limit
            The max possible size.

        Returns
        -------
        zoom
            The valid zoom level.
        """  # numpydoc ignore=ES01
        # find all non-trivial divisors of x
        divisors = [i for i in range(2, x) if not x % i]
        # find the largest zoom (equivalently, smallest divisor) such that the
        # zoomed in image is smaller than the limit
        return 1 / min([i for i in divisors if x / i <= limit])

    if ax.bbox.height > max(image_heights):
        zoom = ax.bbox.height // max(image_heights)
    else:
        zoom = find_zoom_helper(max(image_heights), ax.bbox.height)
    if ax.bbox.width > max(image_widths):
        zoom = min(zoom, ax.bbox.width // max(image_widths))
    else:
        zoom = find_zoom_helper(max(image_widths), ax.bbox.width)
    return zoom


def imshow(
    image: torch.Tensor | list[torch.Tensor],
    vrange: tuple[float, float] | str = "indep1",
    zoom: float | None = None,
    title: str | list[str] | None = "",
    col_wrap: int | None = None,
    ax: mpl.axes.Axes | None = None,
    cmap: mpl.colors.Colormap | None = None,
    plot_complex: Literal["rectangular", "polar", "logpolar"] = "rectangular",
    batch_idx: int | None = None,
    channel_idx: int | None = None,
    as_rgb: bool = False,
    **kwargs: Any,
) -> pt.tools.display.PyrFigure:
    """
    Show image(s), avoiding interpolation.

    This function shows images carefully, avoiding interpolation: each element in the
    input ``image`` will correspond to a pixel or an integer number of pixels. When
    ``zoom<1``, an integer number of input elements will be averaged into a single
    pixel.

    Parameters
    ----------
    image
        The images to display. Tensors should be 4d (batch, channel, height,
        width). List of tensors should be used for tensors of different height
        and width: all images will automatically be rescaled so they're
        displayed at the same height and width, thus, their heights and widths
        must be scalar multiples of each other.
    vrange
        If a 2-tuple, specifies the image values vmin/vmax that are mapped to
        the minimum and maximum value of the colormap, respectively. If a
        string:

        * ``"auto0"``: All images have same vmin/vmax, which have the same
          absolute value, and come from the minimum or maximum across all
          images, whichever has the larger absolute value.

        * ``"auto1"``: All images have same vmin/vmax, which are the
          minimum/maximum values across all images.

        * ``"auto2"``: All images have same vmin/vmax, which are the mean (across
          all images) minus/ plus 2 std dev (across all images).

        * ``"auto3"``: All images have same vmin/vmax, chosen so as to map the 10th/90th
          percentile values to the 10th/90th percentile of the display intensity range.
          For example: vmin is the 10th percentile image value minus 1/8 times the
          difference between the 90th and 10th percentile.

        * ``"indep0"``: Each image has an independent vmin/vmax, which have the same
          absolute value, which comes from either their minimum or maximum value,
          whichever has the larger absolute value.

        * ``"indep1"``: Each image has an independent vmin/vmax, which are their
          minimum/maximum values.

        * ``"indep2"``: Each image has an independent vmin/vmax, which is their
          mean minus/plus 2 std dev.

        * ``"indep3"``: Each image has an independent vmin/vmax, chosen so that the
          10th/90th percentile values map to the 10th/90th percentile intensities.

    zoom
        Ratio of display pixels to image pixels. If greater than 1, must be an
        integer. If less than 1, must be ``1/d`` where ``d`` is a a divisor of the
        size of the largest image. If ``None``, we try to determine the best zoom.
    title
        Title for the plot. In addition to the specified title, we add a
        subtitle giving the plotted range and dimensionality (with zoom).

        * If ``str``, will put the same title on every plot.

        * If ``list``, all values must be ``str``, must be the same length as img,
          and each title will be assigned to corresponding plot.

        * If ``None``, no title will be printed and subtitle will be removed.

    col_wrap
        Number of axes to have in each row. If ``None``, will fit all axes in a
        single row.
    ax
        If ``None``, we make the appropriate figure. Otherwise, we resize the axes
        so that it's the appropriate number of pixels (done by shrinking the
        bbox - if the bbox is already too small, this will throw an Exception,
        so first define a large enough figure).
    cmap
        Colormap to use when showing these images. If ``None``, then behavior is
        determined by ``vrange``: if ``vmap in ["auto0", "indep0"]``, we use
        ``"RdBu_r"``, else we use ``"gray"`` (see `matplotlib documentation
        <https://matplotlib.org/stable/users/explain/colors/colormaps.html#colormaps>`_).
    plot_complex
        Specifies handling of complex values.

        * ``"rectangular"``: plot real and imaginary components as separate images.

        * ``"polar"``: plot amplitude and phase as separate images.

        * ``"logpolar"``: plot log (base 2) amplitude and phase as separate images.

    batch_idx
        Which element from the batch dimension to plot. If ``None``, we plot all.
    channel_idx
        Which element from the channel dimension to plot. If ``None``, we plot all.
        Note if this is not ``None``, then ``as_rgb=True`` will fail, because we
        restrict the channels.
    as_rgb
        Whether to consider the channels as encoding RGB(A) values. If ``True``, we
        attempt to plot the image in color, so your tensor must have 3 (or 4 if
        you want the alpha channel) elements in the channel dimension. If ``False``,
        we plot each channel as a separate grayscale image.
    **kwargs
        Passed to :func:`matplotlib.pyplot.imshow`.

    Returns
    -------
    fig
        Figure containing the plotted images.

    Raises
    ------
    TypeError
        If ``batch_idx`` or ``channel_idx`` takes an illegal value.
    ValueError
        If ``zoom`` takes an illegal value.
    ValueError
        If ``as_rgb=True`` and the input ``image`` does not have 3 or 4 channels.
    ValueError
        If ``as_rgb=False``, ``image`` has more than one channel and one more than
        one batch and neither ``batch_idx`` nor ``channel_idx`` is set.
    Exception
        If ``plot_complex`` takes an illegal value.

    Notes
    -----
    This interpolation avoidance is only guaranteed for the saved image; it should
    generally hold in notebooks as well, but will fail if, e.g., you plot an image
    that's 2000 pixels wide on an monitor 1000 pixels wide; the browser handles the
    rescaling in a way we can't control.
    """
    if not isinstance(image, list):
        image = [image]
    images_to_plot = []
    heights, widths = [], []
    for im in image:
        im = to_numpy(im)
        if im.shape[0] > 1 and batch_idx is not None:
            try:
                # this preserves the number of dimensions
                im = im[batch_idx : batch_idx + 1]
            except TypeError:
                raise TypeError(f"batch_idx must be an int or None but got {batch_idx}")
        if channel_idx is not None:
            try:
                # this preserves the number of dimensions
                im = im[:, channel_idx : channel_idx + 1]
            except TypeError:
                raise TypeError(
                    f"channel_idx must be an int or None but got {channel_idx}"
                )
        # allow RGB and RGBA
        if as_rgb:
            if im.shape[1] not in [3, 4]:
                raise ValueError(
                    "If as_rgb is True, then channel must have 3 or 4 elements!"
                )
            im = im.transpose(0, 2, 3, 1)
            # want to insert a fake "channel" dimension here, so our putting it
            # into a list below works as expected
            im = im.reshape((im.shape[0], 1, *im.shape[1:]))
        elif im.shape[1] > 1 and im.shape[0] > 1:
            raise ValueError(
                "Don't know how to plot non-rgb images with more than one channel"
                " and batch! Use batch_idx / channel_idx to choose a subset for"
                " plotting."
            )
        # by iterating through it twice, we make sure to peel apart the batch
        # and channel dimensions so that they each show up as a separate image.
        # because of how we've handled everything above, we know that im will
        # be (b,c,h,w) or (b,c,h,w,r) where r is the RGB(A) values
        for i in im:
            # at this point, i_ are all shape (h,w) or (h,w,r) and so we don't
            # squeeze, which could accidentally drop a dimension if h or w is a
            # singleton dimension
            images_to_plot.extend([i_ for i_ in i])
            heights.extend([i_.shape[0] for i_ in i])
            widths.extend([i_.shape[1] for i_ in i])

    if zoom is None and ax is not None:
        zoom = _find_zoom(heights, widths, ax)
    elif zoom is None:
        zoom = 1
    elif zoom <= 0:
        raise ValueError("zoom must be positive!")

    return pt.imshow(
        images_to_plot,
        vrange=vrange,
        zoom=zoom,
        title=title,
        col_wrap=col_wrap,
        ax=ax,
        cmap=cmap,
        plot_complex=plot_complex,
        **kwargs,
    )


def animshow(
    video: torch.Tensor | list[torch.Tensor],
    framerate: float = 2.0,
    repeat: bool = False,
    vrange: tuple[float, float] | str = "indep1",
    zoom: float | None = None,
    title: str | list[str] | None = "",
    col_wrap: int | None = None,
    ax: mpl.axes.Axes | None = None,
    cmap: mpl.colors.Colormap | None = None,
    plot_complex: Literal["rectangular", "polar", "logpolar"] = "rectangular",
    batch_idx: int | None = None,
    channel_idx: int | None = None,
    as_rgb: bool = False,
    **kwargs: Any,
) -> mpl.animation.FuncAnimation:
    """
    Animate video(s), avoiding interpolation.

    This function shows images carefully, avoiding interpolation: each element in the
    input ``image`` will correspond to a pixel or an integer number of pixels. When
    ``zoom<1``, an integer number of input elements will be averaged into a single
    pixel.

    This functions returns a matplotlib FuncAnimation object. See our
    documentation (e.g., :ref:`quickstart-nb`) for examples on how to view it in
    a Jupyter notebook. In order to save, use ``anim.save(filename)``. In either
    case, this can take a while and you'll need the appropriate writer installed
    and on your path, e.g., ffmpeg, imagemagick, etc). See
    :doc:`matplotlib documentation <matplotlib:api/animation_api>` for more details.

    Parameters
    ----------
    video
        The video(s) to display. Tensors should be 5d (batch, channel, time,
        height, width). List of tensors should be used for tensors of different
        height and width: all videos will automatically be rescaled so they're
        displayed at the same height and width, thus, their heights and widths
        must be scalar multiples of each other. Videos must all have the same
        number of frames.
    framerate
        Temporal resolution of the video, in Hz (frames per second).
    repeat
        Whether to loop the animation or just play it once.
    vrange
        If a 2-tuple, specifies the image values vmin/vmax that are mapped to
        the minimum and maximum value of the colormap, respectively. If a
        string:

        * ``"auto0"``: All images have same vmin/vmax, which have the same
          absolute value, and come from the minimum or maximum across all
          images, whichever has the larger absolute value.

        * ``"auto1"``: All images have same vmin/vmax, which are the
          minimum/maximum values across all images.

        * ``"auto2"``: All images have same vmin/vmax, which are the mean (across
          all images) minus/ plus 2 std dev (across all images).

        * ``"auto3"``: All images have same vmin/vmax, chosen so as to map the 10th/90th
          percentile values to the 10th/90th percentile of the display intensity range.
          For example: vmin is the 10th percentile image value minus 1/8 times the
          difference between the 90th and 10th percentile.

        * ``"indep0"``: Each image has an independent vmin/vmax, which have the same
          absolute value, which comes from either their minimum or maximum value,
          whichever has the larger absolute value.

        * ``"indep1"``: Each image has an independent vmin/vmax, which are their
          minimum/maximum values.

        * ``"indep2"``: Each image has an independent vmin/vmax, which is their
          mean minus/plus 2 std dev.

        * ``"indep3"``: Each image has an independent vmin/vmax, chosen so that the
          10th/90th percentile values map to the 10th/90th percentile intensities.

    zoom
        Ratio of display pixels to image pixels. If greater than 1, must be an
        integer. If less than 1, must be ``1/d`` where ``d`` is a a divisor of the
        size of the largest image. If ``None``, we try to determine the best zoom.
    title
        Title for the plot. In addition to the specified title, we add a
        subtitle giving the plotted range and dimensionality (with zoom).

        * If ``str``, will put the same title on every plot.

        * If ``list``, all values must be ``str``, must be the same length as img,
          and each title will be assigned to corresponding plot.

        * If ``None``, no title will be printed and subtitle will be removed.

    col_wrap
        Number of axes to have in each row. If ``None``, will fit all axes in a
        single row.
    ax
        If ``None``, we make the appropriate figure. Otherwise, we resize the axes
        so that it's the appropriate number of pixels (done by shrinking the
        bbox - if the bbox is already too small, this will throw an Exception,
        so first define a large enough figure).
    cmap
        Colormap to use when showing these images. If ``None``, then behavior is
        determined by ``vrange``: if ``vmap in ["auto0", "indep0"]``, we use
        ``"RdBu_r"``, else we use ``"gray"`` (see `matplotlib documentation
        <https://matplotlib.org/stable/users/explain/colors/colormaps.html#colormaps>`_).
    plot_complex
        Specifies handling of complex values.

        * ``"rectangular"``: plot real and imaginary components as separate images.

        * ``"polar"``: plot amplitude and phase as separate images.

        * ``"logpolar"``: plot log (base 2) amplitude and phase as separate images.

    batch_idx
        Which element from the batch dimension to plot. If ``None``, we plot all.
    channel_idx
        Which element from the channel dimension to plot. If ``None``, we plot all.
        Note if this is not ``None``, then ``as_rgb=True`` will fail, because we
        restrict the channels.
    as_rgb
        Whether to consider the channels as encoding RGB(A) values. If ``True``, we
        attempt to plot the image in color, so your tensor must have 3 (or 4 if
        you want the alpha channel) elements in the channel dimension. If ``False``,
        we plot each channel as a separate grayscale image.
    **kwargs
        Passed to :func:`matplotlib.pyplot.imshow`.

    Returns
    -------
    anim
        The animation object. In order to view, must convert to HTML
        or save.

    Raises
    ------
    TypeError
        If ``batch_idx`` or ``channel_idx`` takes an illegal value.
    ValueError
        If ``as_rgb=True`` and the input ``image`` does not have 3 or 4 channels.
    ValueError
        If ``as_rgb=False``, ``image`` has more than one channel and one more than
        one batch and neither ``batch_idx`` nor ``channel_idx`` is set.
    Exception
        If ``plot_complex`` takes an illegal value.

    Notes
    -----
    - Unless specified, we use the ffmpeg backend, which requires that you have
      ffmpeg installed and on your path (https://ffmpeg.org/download.html). To
      use a different, use the matplotlib rcParams:
      ``matplotlib.rcParams['animation.writer'] = writer``, see `matplotlib
      documentation
      <https://matplotlib.org/stable/api/animation_api.html#writer-classes>`_
      for more details.

    - This interpolation avoidance is only guaranteed for the saved image; it should
      generally hold in notebooks as well, but will fail if, e.g., you plot an image
      that's 2000 pixels wide on an monitor 1000 pixels wide; the browser handles the
      rescaling in a way we can't control.
    """
    if not isinstance(video, list):
        video = [video]
    videos_to_show = []
    heights, widths = [], []
    for vid in video:
        vid = to_numpy(vid)
        if vid.shape[0] > 1 and batch_idx is not None:
            # this preserves the number of dimensions
            try:
                vid = vid[batch_idx : batch_idx + 1]
            except TypeError:
                raise TypeError(f"batch_idx must be an int or None but got {batch_idx}")
        if channel_idx is not None:
            try:
                # this preserves the number of dimensions
                vid = vid[:, channel_idx : channel_idx + 1]
            except TypeError:
                raise TypeError(
                    f"channel_idx must be an int or None but got {channel_idx}"
                )
        # allow RGB and RGBA
        if as_rgb:
            if vid.shape[1] not in [3, 4]:
                raise ValueError(
                    "If as_rgb is True, then channel must have 3 or 4 elements!"
                )
            vid = vid.transpose(0, 2, 3, 4, 1)
            # want to insert a fake "channel" dimension here, so our putting it
            # into a list below works as expected
            vid = vid.reshape((vid.shape[0], 1, *vid.shape[1:]))
        elif vid.shape[1] > 1 and vid.shape[0] > 1:
            raise ValueError(
                "Don't know how to plot non-rgb images with more than one channel and"
                " batch! Use batch_idx / channel_idx to choose a subset for"
                " plotting"
            )
        # by iterating through it twice, we make sure to peel apart the batch
        # and channel dimensions so that they each show up as a separate video.
        # because of how we've handled everything above, we know that vid will
        # be (b,c,t,h,w) or (b,c,t,h,w,r) where r is the RGB(A) values
        for v in vid:
            videos_to_show.extend([v_.squeeze() for v_ in v])
            heights.extend([v_.shape[1] for v_ in v])
            widths.extend([v_.shape[2] for v_ in v])

    if zoom is None and ax is not None:
        zoom = _find_zoom(heights, widths, ax)
    elif zoom is None:
        zoom = 1
    elif zoom <= 0:
        raise ValueError("zoom must be positive!")

    return pt.animshow(
        videos_to_show,
        framerate=framerate,
        as_html5=False,
        repeat=repeat,
        vrange=vrange,
        zoom=zoom,
        title=title,
        col_wrap=col_wrap,
        ax=ax,
        cmap=cmap,
        plot_complex=plot_complex,
        **kwargs,
    )


def pyrshow(
    pyr_coeffs: dict,
    vrange: tuple[float, float] | str = "indep1",
    zoom: float = 1,
    show_residuals: bool = True,
    cmap: mpl.colors.Colormap | None = None,
    plot_complex: Literal["rectangular", "polar", "logpolar"] = "rectangular",
    batch_idx: int = 0,
    channel_idx: int = 0,
    **kwargs: Any,
) -> pt.tools.display.PyrFigure:
    r"""
    Display steerable pyramid coefficients in orderly fashion.

    This function uses :func:`~plenoptic.tools.display.imshow` to show the coefficients
    of the steeable pyramid (or any dictionary in the standard format),
    such that each scale shows up on a single row, with each scale in a given column.

    Note that unlike :func:`~plenoptic.tools.display.imshow`, we can only show one batch
    or channel at a time.

    Parameters
    ----------
    pyr_coeffs
        Pyramid coefficients in the standard dictionary format as returned by
        ``SteerablePyramidFreq.forward()``.
    vrange
        If a 2-tuple, specifies the image values vmin/vmax that are mapped to
        the minimum and maximum value of the colormap, respectively. If a
        string:

        * ``"auto0"``: All images have same vmin/vmax, which have the same
          absolute value, and come from the minimum or maximum across all
          images, whichever has the larger absolute value.

        * ``"auto1"``: All images have same vmin/vmax, which are the
          minimum/maximum values across all images.

        * ``"auto2"``: All images have same vmin/vmax, which are the mean (across
          all images) minus/ plus 2 std dev (across all images).

        * ``"auto3"``: All images have same vmin/vmax, chosen so as to map the 10th/90th
          percentile values to the 10th/90th percentile of the display intensity range.
          For example: vmin is the 10th percentile image value minus 1/8 times the
          difference between the 90th and 10th percentile.

        * ``"indep0"``: Each image has an independent vmin/vmax, which have the same
          absolute value, which comes from either their minimum or maximum value,
          whichever has the larger absolute value.

        * ``"indep1"``: Each image has an independent vmin/vmax, which are their
          minimum/maximum values.

        * ``"indep2"``: Each image has an independent vmin/vmax, which is their
          mean minus/plus 2 std dev.

        * ``"indep3"``: Each image has an independent vmin/vmax, chosen so that the
          10th/90th percentile values map to the 10th/90th percentile intensities.

    zoom
        Ratio of display pixels to image pixels. if greater than 1, must be an
        integer. If less than 1, must be ``1/d`` where ``d`` is a a divisor of the
        size of the largest image.
    show_residuals
        Whether to display the residual bands.
    cmap
        Colormap to use when showing these images.
    plot_complex : {'rectangular', 'polar', 'logpolar'}
        Specifies handling of complex values.

        * ``"rectangular"``: plot real and imaginary components as separate images.

        * ``"polar"``: plot amplitude and phase as separate images.

        * ``"logpolar"``: plot log (base 2) amplitude and phase as separate images.
    batch_idx
        Which element from the batch dimension to plot.
    channel_idx
        Which element from the channel dimension to plot.
    **kwargs
        Passed on to :func:`pyrtools.tools.display.pyrshow`.

    Returns
    -------
    fig
        The figure displaying the coefficients.

    Raises
    ------
    TypeError
        If ``batch_idx`` or ``channel_idx`` takes an illegal value.
    """
    pyr_coeffvis = {}
    is_complex = False
    for k, v in pyr_coeffs.items():
        im = to_numpy(v)
        if np.iscomplex(im).any():
            is_complex = True
        try:
            # this removes only the first (batch) dimension
            im = im[batch_idx : batch_idx + 1].squeeze(0)
        except TypeError:
            raise TypeError(f"batch_idx must be an int but got {batch_idx}")
        try:
            # this removes only the first (now channel) dimension
            im = im[channel_idx : channel_idx + 1].squeeze(0)
        except TypeError:
            raise TypeError(f"channel_idx must be an int but got {channel_idx}")
        # because of how we've handled everything above, we know that im will
        # be (h,w).
        pyr_coeffvis[k] = im

    return pt.pyrshow(
        pyr_coeffvis,
        is_complex=is_complex,
        vrange=vrange,
        zoom=zoom,
        cmap=cmap,
        plot_complex=plot_complex,
        show_residuals=show_residuals,
        **kwargs,
    )


def clean_up_axes(
    ax: mpl.axes.Axes,
    ylim: tuple[float, float] | None | Literal[False] = None,
    spines_to_remove: list[Literal["top", "right", "bottom", "left"]] = [
        "top",
        "right",
        "bottom",
    ],
    axes_to_remove: list[Literal["x", "y"]] = ["x"],
) -> mpl.axes.Axes:
    r"""
    Clean up an axis, as desired when making a stem plot of the representation.

    This function can:

    - Remove the spines from axis (the axis lines and tick marks).

    - Set axis objects themselves invisible (includes not just spines but also tick
      labels and axis label).

    - Set ylim.

    Parameters
    ----------
    ax
        The axis to clean up.
    ylim
        If a tuple, the y-limits to use for this plot. If ``None``, we use
        the original limits, slightly adjusted so that the minimum is 0. If
        ``False``, we do not change y-limits.
    spines_to_remove
        The spines to remove from the axis.
    axes_to_remove
        The axes to set as invisible.

    Returns
    -------
    ax
        The cleaned-up axis.
    """
    if spines_to_remove is None:
        spines_to_remove = ["top", "right", "bottom"]
    if axes_to_remove is None:
        axes_to_remove = ["x"]

    if ylim is not None:
        if ylim:
            ax.set_ylim(ylim)
    else:
        ax.set_ylim((0, ax.get_ylim()[1]))
    if "x" in axes_to_remove:
        ax.xaxis.set_visible(False)
    if "y" in axes_to_remove:
        ax.yaxis.set_visible(False)
    for s in spines_to_remove:
        ax.spines[s].set_visible(False)
    return ax


def update_stem(
    stem_container: mpl.container.StemContainer, ydata: np.ndarray | torch.Tensor
) -> mpl.container.StemContainer:
    r"""
    Update the information in a stem plot.

    We update the information in a single stem plot to match that given
    by ``ydata``. We update the position of the markers and and the
    lines connecting them to the baseline, but we don't change the
    baseline at all and assume that the xdata shouldn't change at all.

    Parameters
    ----------
    stem_container
        Single container for the artists created in a :func:`matplotlib.pyplot.stem`
        plot. It can be treated like a namedtuple ``(markerline, stemlines, baseline)``.
        In order to get this from an axis ``ax``, try ``ax.containers[0]`` (if you have
        more than one container in that axis, it may not be the first one).
    ydata
        The new y-data to show on the plot. Importantly, must be the same length as
        the existing y-data.

    Returns
    -------
    stem_container
        The StemContainer containing the updated artists.
    """
    stem_container.markerline.set_ydata(ydata)
    segments = stem_container.stemlines.get_segments().copy()
    for s, y in zip(segments, ydata):
        try:
            s[1, 1] = y
        except IndexError:
            # this happens when our segment array is 1x2 instead of 2x2,
            # which is the case when the data there is nan
            continue
    stem_container.stemlines.set_segments(segments)
    return stem_container


def rescale_ylim(axes: list[mpl.axes.Axes], data: np.ndarray | torch.Tensor):
    r"""
    Rescale y-limits nicely.

    We take the axes and set their limits to be ``(-y_max, y_max)``,
    where ``y_max=np.abs(data).max()``.

    Parameters
    ----------
    axes
        A list of matplotlib axes to rescale.
    data
        The data to use when rescaling (or a dictionary of such values).
    """
    data = data.cpu()

    def find_ymax(data: np.ndarray | torch.Tensor) -> float:
        """
        Find appropriate ymax.

        Parameters
        ----------
        data
            The tensor whose ymax we should grab.

        Returns
        -------
        ymax
            The appropriate ymax.
        """  # numpydoc ignore=ES01
        try:
            return np.abs(data).max()
        except RuntimeError:
            # then we need to call to_numpy on it because it needs to be
            # detached and converted to an array
            return np.abs(to_numpy(data)).max()

    try:
        y_max = find_ymax(data)
    except TypeError:
        # then this is a dictionary
        y_max = np.max([find_ymax(d) for d in data.values()])
    for ax in axes:
        ax.set_ylim((-y_max, y_max))


def clean_stem_plot(
    data: np.ndarray,
    ax: mpl.axes.Axes | None = None,
    title: str | None = "",
    ylim: tuple | None | Literal[False] = None,
    xvals: tuple[list[float], list[float]] | None = None,
    **kwargs: Any,
) -> mpl.axes.Axes:
    r"""
    Create a simple stem plot.

    This plots the data, baseline, cleans up the axis, and sets the
    title.

    Helper function for :func:`~plenoptic.tools.display.plot_representation()`.

    If ``xvals=None``, stem plot will have a baseline that covers the entire range
    of the data. In order to break that up visually (so there's a line from 0 to 9,
    from 10 to 19, etc) pass ``xvals`` separately.

    Parameters
    ----------
    data
        The data to plot (as a stem plot).
    ax
        The axis to plot the data on. If ``None``, we plot on the current axis
        (grabbed with :func:`matplotlib.pyplot.gca`).
    title
        The title to put on the axis. If ``None``, we don't call ``ax.set_title``
        (useful if you want to avoid changing the title on an existing plot).
    ylim
        If a tuple, the y-limits to use for this plot. If ``None``, we use
        the original limits, slightly adjusted so that the minimum is 0. If
        ``False``, we do not change y-limits.
    xvals
        A 2-tuple of lists, containing the start (``xvals[0]``) and stop
        (``xvals[1]``) x values for plotting. If ``None``, baseline will cover
        full range.
    **kwargs
        Passed to :func:`matplotlib.pyplot.stem`.

    Returns
    -------
    ax
        The axis with the plot.

    Examples
    --------
    We allow for breaks in the baseline value if we want to visually
    break up the plot, as we see below.

    .. plot::

      >>> import plenoptic as po
      >>> import numpy as np
      >>> import matplotlib.pyplot as plt
      >>> # if ylim=None, as in this example, the minimum y-valuewill get
      >>> # set to 0, so we want to make sure our values are all positive
      >>> y = np.abs(np.random.randn(55))
      >>> y[15:20] = np.nan
      >>> y[35:40] = np.nan
      >>> # we want to draw the baseline from 0 to 14, 20 to 34, and 40 to
      >>> # 54, everywhere that we have non-NaN values for y
      >>> xvals = ([0, 20, 40], [14, 34, 54])
      >>> po.tools.display.clean_stem_plot(y, xvals=xvals)
      <Axes: >

    If we don't care about breaking up the x-axis, you can set ``xvals=None``.
    In this case, this function will just clean up the plot a little bit.

    .. plot::

      >>> import plenoptic as po
      >>> import numpy as np
      >>> import matplotlib.pyplot as plt
      >>> # if ylim=None, as in this example, the minimum y-valuewill get
      >>> # set to 0, so we want to make sure our values are all positive
      >>> y = np.abs(np.random.randn(55))
      >>> po.tools.display.clean_stem_plot(y)
      <Axes: >
    """
    if ax is None:
        ax = plt.gca()
    if xvals is not None:
        basefmt = " "
        ax.hlines(len(xvals[0]) * [0], xvals[0], xvals[1], colors="C3", zorder=10)
    else:
        # this is the default basefmt value
        basefmt = None
    ax.stem(data, basefmt=basefmt, **kwargs)
    ax = clean_up_axes(ax, ylim, ["top", "right", "bottom"])
    if title is not None:
        ax.set_title(title)
    return ax


def _get_artists_from_axes(
    axes: mpl.axes.Axes | list[mpl.axes.Axes],
    data: torch.Tensor | dict,
) -> dict:
    """
    Grab artists from axes.

    For now, we only grab containers (stem plots), images, or lines

    See the docstring of :meth:`~plenoptic.tools.display.update_plot()` for details on
    how ``axes`` and ``data`` should be structured.

    Parameters
    ----------
    axes
        The axis/axes to update.
    data
        The new data to plot.

    Returns
    -------
    artists
        Dictionary of artists for updating plots. Values are the artists to
        use, keys are the corresponding keys from data.

    Raises
    ------
    ValueError
        If the number of artists in ``axes`` is different from the size of the
        dimension-to-plot of ``data``.
    """
    if not hasattr(axes, "__iter__"):
        # then we only have one axis, so we may be able to update more than one
        # data element.
        if len(axes.containers) > 0:
            data_check = 1
            artists = axes.containers
        elif len(axes.images) > 0:
            # images are weird, so don't check them like this
            data_check = None
            artists = axes.images
        elif len(axes.lines) > 0:
            data_check = 1
            artists = axes.lines
        elif len(axes.collections) > 0:
            data_check = 2
            artists = axes.collections
        if isinstance(data, dict):
            artists = {ax.get_label(): ax for ax in artists}
        else:
            if data_check == 1 and data.shape[1] != len(artists):
                raise ValueError(
                    f"data has {data.shape[1]} things to plot, but "
                    f"your axis contains {len(artists)} plotting artists, "
                    "so unsure how to continue! Pass data as a dictionary"
                    " with keys corresponding to the labels of the artists"
                    " to update to resolve this."
                )
            elif data_check == 2 and data.ndim > 2 and data.shape[-3] != len(artists):
                raise ValueError(
                    f"data has {data.shape[-3]} things to plot, but "
                    f"your axis contains {len(artists)} plotting artists, "
                    "so unsure how to continue! Pass data as a dictionary"
                    " with keys corresponding to the labels of the artists"
                    " to update to resolve this."
                )
    else:
        # then we have multiple axes, so we are only updating one data element
        # per plot
        artists = []
        for ax in axes:
            if len(ax.containers) == 1:
                data_check = 1
                artists.extend(ax.containers)
            elif len(ax.images) == 1:
                # images are weird, so don't check them like this
                data_check = None
                artists.extend(ax.images)
            elif len(ax.lines) == 1:
                artists.extend(ax.lines)
                data_check = 1
            elif len(ax.collections) == 1:
                artists.extend(ax.collections)
                data_check = 2
        if isinstance(data, dict):
            if len(data.keys()) != len(artists):
                raise ValueError(
                    f"data has {len(data.keys())} things to plot, but "
                    f"you passed {len(axes)} axes , so unsure how "
                    "to continue!"
                )
            artists = {k: a for k, a in zip(data.keys(), artists)}
        else:
            if data_check == 1 and data.shape[1] != len(artists):
                raise ValueError(
                    f"data has {data.shape[1]} things to plot, but "
                    f"you passed {len(axes)} axes , so unsure how "
                    "to continue!"
                )
            if data_check == 2 and data.ndim > 2 and data.shape[-3] != len(artists):
                raise ValueError(
                    f"data has {data.shape[-3]} things to plot, but "
                    f"you passed {len(axes)} axes , so unsure how "
                    "to continue!"
                )
    if not isinstance(artists, dict):
        artists = {f"{i:02d}": a for i, a in enumerate(artists)}
    return artists


def update_plot(
    axes: mpl.axes.Axes | list[mpl.axes.Axes],
    data: torch.Tensor | dict,
    model: torch.nn.Module | None = None,
    batch_idx: int = 0,
) -> list:
    r"""
    Update the information in some axes.

    This is used for creating an animation over time. In order to create the animation,
    we need to know how to update the matplotlib Artists, and this provides a simple way
    of doing that. It assumes the plot has been created by something like
    :func:`~plenoptic.tools.display.plot_representation`, which initializes all the
    artists.

    We can update stem plots, lines (as returned by :func:`matplotlib.pyplot.plot`),
    scatter plots, or images (RGB, RGBA, or grayscale).

    There are two modes for this:

    - Single axis: ``axes`` is a single axis, which may contain multiple artists (all of
      the same type) to update. ``data`` should be a :class:`torch.Tensor` with multiple
      channels (one per artist in the same order) or be a dictionary whose keys give the
      label(s) of the corresponding artist(s) and whose values are
      :class:`torch.Tensor`.

    - Multiple axes: ``axes`` is a list of axes, each of which contains a single artist
      to update (artists can be different types). ``data`` should be a
      :class:`torch.Tensor` with multiple channels (one per axis in the same order) or a
      dictionary with the same number of keys as ``axes``, which we can iterate through
      in order, and whose values are :class:`torch.Tensor`.

    In all cases, ``data`` Tensors should be 3d (if the plot we're updating is a
    line or stem plot) or 4d (if it's an image or scatter plot).

    RGB(A) images are special, since we store that info along the channel
    dimension, so they only work with single-axis mode (which will only have a
    single artist, because that's how imshow works).

    If you have multiple axes, each with multiple artists you want to update, that's too
    complicated for us, and so you should write a ``model.update_plot()`` function which
    handles that (see
    :func:`plenoptic.simulate.models.portilla_simoncelli.PortillaSimoncelli.update_plot`
    for an example).

    If ``model`` is set, we try to call ``model.update_plot()`` (which must also return
    artists). If ``model`` doesn't have an ``update_plot`` method, then we try to figure
    out how to update the axes ourselves, based on the shape of the data.

    Parameters
    ----------
    axes
        The axis or list of axes to update. We assume that these are the axes created by
        :func:`~plenoptic.tools.display.plot_representation` and so contain artists
        in the correct order.
    data
        The new data to plot.
    model
        A differentiable model that tells us how to plot ``data``. See above for
        behavior if ``None``.
    batch_idx
        Which index to take from the batch dimension.

    Returns
    -------
    artists
        A list of the artists used to update the information on the
        plots.

    Raises
    ------
    ValueError
        If ``data`` (or its values, if it's a ``dict``) are not 3 or 4 dimensional.
    """
    if isinstance(data, dict):
        for v in data.values():
            if v.ndim not in [3, 4]:
                raise ValueError(
                    "update_plot expects 3 or 4 dimensional data"
                    "; unexpected behavior will result otherwise!"
                    f" Got data of shape {v.shape}"
                )
    else:
        if data.ndim not in [3, 4]:
            raise ValueError(
                "update_plot expects 3 or 4 dimensional data"
                "; unexpected behavior will result otherwise!"
                f" Got data of shape {data.shape}"
            )
    try:
        artists = model.update_plot(axes=axes, batch_idx=batch_idx, data=data)
    except AttributeError:
        ax_artists = _get_artists_from_axes(axes, data)
        artists = []
        if not isinstance(data, dict):
            data_dict = {}
            # check for RGBA images
            if len(ax_artists) == 1 and data.shape[1] > 1:
                # can't index into dict.values(), so use this work around
                # instead, as suggested
                # https://stackoverflow.com/questions/43629270/how-to-get-single-value-from-dict-with-single-entry
                try:
                    if next(iter(ax_artists.values())).get_array().data.ndim > 1:
                        # then this is an RGBA image
                        data_dict = {"00": data}
                except Exception as e:
                    raise ValueError(
                        "Thought this was an RGB(A) image based on the number"
                        " of artists and data shape, but something is off!"
                        f" Original exception: {e}"
                    )
            else:
                for i, d in enumerate(data.unbind(1)):
                    # need to keep the shape the same because of how we
                    # check for shape below (unbinding removes a dimension,
                    # so we add it back)
                    data_dict[f"{i:02d}"] = d.unsqueeze(1)
            data = data_dict
        for k, d in data.items():
            try:
                art = ax_artists[k]
            except KeyError:
                # If the we're grabbing these labels from the line labels and
                # they were originally ints, they will get converted to
                # strings. this catches that
                art = ax_artists[str(k)]
            d = to_numpy(d[batch_idx]).squeeze()
            if d.ndim == 1:
                try:
                    # then it's a line
                    x, _ = art.get_data()
                    art.set_data(x, d)
                    artists.append(art)
                except AttributeError:
                    # then it's a stemplot
                    sc = update_stem(art, d)
                    artists.extend([sc.markerline, sc.stemlines])
            elif d.ndim == 2:
                try:
                    # then it's a grayscale image
                    art.set_data(d)
                    artists.append(art)
                except AttributeError:
                    # then it's a scatterplot
                    art.set_offsets(d)
                    artists.append(art)
            else:
                # then it's an RGB(A) image. for tensors, we put that dimension
                # in channel, but for images, it should be at the end
                art.set_data(np.moveaxis(d, 0, -1))
                artists.append(art)
    # make sure to always return a list
    if not isinstance(artists, list):
        artists = [artists]
    return artists


def plot_representation(
    model: torch.nn.Module | None = None,
    data: np.ndarray | dict | None = None,
    ax: mpl.axes.Axes | None = None,
    figsize: tuple[float, float] | None = None,
    ylim: tuple[float, float] | None | Literal[False] = False,
    batch_idx: int = 0,
    title: str = "",
    as_rgb: bool = False,
) -> list[mpl.axes.Axes]:
    r"""
    Plot model representation.

    We try to plot ``data`` on ``ax``, using the ``model.plot_representation`` method,
    if it has it, and otherwise use a function that makes sense based on the
    shape of ``data``.

    All of these arguments are optional, but at least some of them need
    to be set:

    - If ``model`` is ``None``, we fall-back to a type of plot based on the shape of
      ``data``. If it looks image-like, we'll use
      :func:`~plenoptic.tools.display.imshow` and if it looks vector-like, we'll use
      :func:`~plenoptic.tools.display.clean_stem_plot`. If it's a dictionary, we'll
      assume each key, value pair gives the title and data to plot on a separate
      sub-plot.

    - If ``data`` is ``None``, we can only do something if
      ``model.plot_representation`` has some behavior when
      ``data=None``; this is probably to plot its own ``representation``
      attribute. Thus, this will raise an Exception if both ``model`` and
      ``data`` are ``None``, because we have no idea what to plot then.

    - If ``ax`` is ``None``, we create a one-subplot figure using ``figsize``.
      If ``ax`` is not ``None``, we therefore ignore ``figsize``.

    - If ``ylim`` is ``None``, we call :func:`~plenoptic.tools.display.rescale_ylim`,
      which sets the axes' y-limits to be ``(-y_max, y_max)``, where
      ``y_max=np.abs(data).max()``. If it's ``False``, we do nothing.

    Parameters
    ----------
    model
        A differentiable model that tells us how to plot ``data``. See
        above for behavior if ``None``.
    data
        The data to plot. See above for behavior if ``None``.
    ax
        The axis to plot on. See above for behavior if ``None``.
    figsize
        The size of the figure to create. Must be ``None`` if ax is not ``None``. If
        both figsize and ax are ``None``, then we use ``figsize=(5, 5)``.
    ylim
        The y-limits to use for this plot. See above for behavior if ``None``.
        If ``False``, we do nothing. Ignored if ``data`` looks image-like.
    batch_idx
        Which index to take from the batch dimension.
    title
        The title to put above this axis. If you want no title, pass
        the empty string (``""``).
    as_rgb
        Whether to consider the channels as encoding RGB(A) values. It will be
        ignored if the representation doesn't look image-like or if the
        model has its own plot_representation_error() method. Else, it will
        be passed to :func:`~plenoptic.tools.display.imshow`, see that method's
        docstring for details.

    Returns
    -------
    axes
        List of created axes.

    Raises
    ------
    ValueError
        If both ``figsize`` and ``ax`` are not ``None``.
    ValueError
        If ``data`` (or its values, if it's a ``dict``) are not 3 or 4 dimensional.
    """
    if ax is None:
        if figsize is None:
            figsize = (5, 5)
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    else:
        if figsize is not None:
            raise ValueError("figsize can't be set if ax is not None")
        fig = ax.figure
    try:
        # no point in passing figsize, because we've already created
        # and are passing an axis or are passing the user-specified one
        fig, axes = model.plot_representation(
            ylim=ylim, ax=ax, title=title, batch_idx=batch_idx, data=data
        )
    except AttributeError:
        if data is None:
            data = model.representation
        if not isinstance(data, dict):
            if title is None:
                title = "Representation"
            data_dict = {}
            if not as_rgb:
                # then we peel apart the channels
                for i, d in enumerate(data.unbind(1)):
                    # need to keep the shape the same because of how we
                    # check for shape below (unbinding removes a dimension,
                    # so we add it back)
                    data_dict[title + f"_{i:02d}"] = d.unsqueeze(1)
            else:
                data_dict[title] = data
            data = data_dict
        else:
            warnings.warn("data has keys, so we're ignoring title!")
        # want to make sure the axis we're taking over is basically invisible.
        ax = clean_up_axes(ax, False, ["top", "right", "bottom", "left"], ["x", "y"])
        axes = []
        if len(list(data.values())[0].shape) == 3:
            # then this is 'vector-like'
            gs = ax.get_subplotspec().subgridspec(
                min(4, len(data)), int(np.ceil(len(data) / 4))
            )
            for i, (k, v) in enumerate(data.items()):
                ax = fig.add_subplot(gs[i % 4, i // 4])
                # only plot the specified batch, but plot each channel
                # in a separate call. there should probably only be one,
                # and if there's not you probably want to do things
                # differently
                for d in v[batch_idx]:
                    ax = clean_stem_plot(to_numpy(d), ax, k, ylim)
                axes.append(ax)
        elif len(list(data.values())[0].shape) == 4:
            # then this is 'image-like'
            gs = ax.get_subplotspec().subgridspec(
                int(np.ceil(len(data) / 4)), min(4, len(data))
            )
            for i, (k, v) in enumerate(data.items()):
                ax = fig.add_subplot(gs[i // 4, i % 4])
                ax = clean_up_axes(
                    ax, False, ["top", "right", "bottom", "left"], ["x", "y"]
                )
                # only plot the specified batch
                imshow(
                    v,
                    batch_idx=batch_idx,
                    title=k,
                    ax=ax,
                    vrange="indep0",
                    as_rgb=as_rgb,
                )
                axes.append(ax)
            # because we're plotting image data, don't want to change
            # ylim at all
            ylim = False
        else:
            raise ValueError(f"Don't know what to do with data of shape {data.shape}")
    if ylim is None:
        if isinstance(data, dict):
            data = torch.cat(list(data.values()), dim=2)
        rescale_ylim(axes, data)
    return axes



"""
Steerable frequency pyramid.

Construct a steerable pyramid on matrix two dimensional signals, in the
Fourier domain.
"""

import warnings
from collections import OrderedDict
from typing import Literal

import numpy as np
import torch
import torch.fft as fft
import torch.nn as nn
from einops import rearrange
from numpy.typing import NDArray
from scipy.special import factorial
from torch import Tensor

from ...tools.signal import interpolate1d, raised_cosine, steer

complex_types = [torch.cdouble, torch.cfloat]

SCALES_TYPE = int | Literal["residual_lowpass", "residual_highpass"]
KEYS_TYPE = tuple[int, int] | Literal["residual_lowpass", "residual_highpass"]


class SteerablePyramidFreq(nn.Module):
    r"""
    Steerable frequency pyramid in Torch.

    Construct a steerable pyramid on matrix two dimensional signals, in the Fourier
    domain. Boundary-handling is circular. Reconstruction is exact (within floating
    point errors). However, if the image has an odd-shape, the reconstruction will not
    be exact due to boundary-handling issues that have not been resolved. Similarly,
    a complex pyramid of order=0 has non-exact reconstruction and cannot be tight-frame.

    The squared radial functions tile the Fourier plane with a raised-cosine
    falloff. Angular functions are

    .. math::

       \cos\left(\frac{\theta-k*\pi}{o+1}\right)^o

    where :math:`o` is the order parameter set at initialization and :math:`k` runs from
    0 to :math:`o` for a total of :math:`o+1` orientations.

    Parameters
    ----------
    image_shape
        Shape of input image.
    height
        The height of the pyramid. If ``'auto'``, will automatically determine based on
        the size of ``image``. If an ``int``, must be non-negative and less than
        ``log2(min(image_shape[1], image_shape[1]))-2``. If ``height=0``, this only
        returns the residuals.
    order
        The Gaussian derivative order used for the steerable filters, in ``[0, 15]``.
        Note that to achieve steerability the minimum number of orientation is
        ``order + 1``, which is used here. To get more orientations at the same order,
        use the method :meth:`steer_coeffs`.
    twidth
        The width of the transition region of the radial lowpass function, in
        octaves.
    is_complex
        Whether the pyramid coefficients should be complex or not. If ``True``, the real
        and imaginary parts correspond to a pair of odd and even symmetric filters. If
        ``False``, the coefficients only include the real part. Regardless of the value
        of ``is_complex``, the symmetry of the real part is determined by the ``order``
        parameter: if ``order`` is even, then the real coefficients are even symmetric;
        if ``order`` is odd, then the real coefficients are odd symmetric. (If
        ``is_complex=True``, then the imaginary coefficients will have the opposite
        symmetry of the real ones).
    downsample
        Whether to downsample each scale in the pyramid or keep the output
        pyramid coefficients in fixed bands of size ``image_shape``. When
        downsample is ``False``, the forward method returns a tensor.
    tight_frame
        Whether the pyramid obeys the generalized parseval theorem or not (i.e.
        is a tight frame). If ``True``, the energy of the pyr_coeffs equals the energy
        of the image. In order to match the `matlabPyrTools
        <http://github.com/labForComputationalVision/matlabpyrtools>`_ or `pyrtools
        <https://github.com/labForComputationalVision/pyrtools>`_ implementations, this
        must be set to ``False``.

    Attributes
    ----------
    image_shape : tuple
        Shape of input image.
    pyr_size : dict
        Dictionary containing the sizes of the pyramid coefficients. Keys are
        ``(level, band)`` tuples and values are tuples.
    fft_norm : str
        The way the ffts are normalized, see :func:`torch.fft.fft2` for more details.
    is_complex : bool
        Whether the coefficients are complex- or real-valued.
    scales : list
        All the scales of the representation (including residuals) in coarse-to-fine
        order. A subset of this list can be passed to the :meth:`forward` method to
        restrict the output.

    Raises
    ------
    ValueError
        If ``image_shape`` contains non-integers.
    ValueError
        If ``len(image_shape) != 2`` .
    ValueError
        If ``height`` is not a non-negative integer or is larger than the biggest
        possible value (determined by ``image_shape``).
    ValueError
        If ``order`` not an integer in ``[0, 15]``.
    ValueError
        If ``order == 0`` and ``is_complex is False``. See
        https://github.com/plenoptic-org/plenoptic/issues/326 for an explanation
    ValueError
        If ``twidth`` not positive.

    Warns
    -----
    UserWarning
        If ``image_shape`` has an odd value, because then reconstruction will be
        imperfect.

    Notes
    -----
    Transform described in [1]_, filter kernel design described in [2]_.
    For further information see the project [3]_.

    References
    ----------
    .. [1] E P Simoncelli and W T Freeman, "The Steerable Pyramid: A Flexible
       Architecture for Multi-Scale Derivative Computation," Second Int'l Conf
       on Image Processing, Washington, DC, Oct 1995.
    .. [2] A Karasaridis and E P Simoncelli, "A Filter Design Technique for
       Steerable Pyramid Image Transforms", ICASSP, Atlanta, GA, May 1996. ..
    .. [3] `<https://www.cns.nyu.edu/~eero/steerpyr/>`_

    Examples
    --------
    >>> import plenoptic as po
    >>> spyr = po.simul.SteerablePyramidFreq((256, 256))
    """

    def __init__(
        self,
        image_shape: tuple[int, int],
        height: Literal["auto"] | int = "auto",
        order: int = 3,
        twidth: int = 1,
        is_complex: bool = False,
        downsample: bool = True,
        tight_frame: bool = False,
    ):
        super().__init__()

        self.pyr_size = OrderedDict()
        self.order = order
        try:
            self.image_shape = tuple([int(i) for i in image_shape])
        except ValueError:
            raise ValueError(
                f"image_shape must be castable to ints, but got {image_shape}!"
            )
        if self.image_shape != tuple(image_shape):
            raise ValueError(
                f"image_shape must be castable to ints, but got {image_shape}!"
            )
        if len(self.image_shape) != 2:
            raise ValueError(
                f"image_shape must be a tuple of length 2, but got {self.image_shape}!"
            )

        if (self.image_shape[0] % 2 != 0) or (self.image_shape[1] % 2 != 0):
            warnings.warn("Reconstruction will not be perfect with odd-sized images")

        self.is_complex = is_complex
        self.downsample = downsample
        self.tight_frame = tight_frame
        if self.tight_frame:
            self.fft_norm = "ortho"
        else:
            self.fft_norm = "backward"
        # cache constants
        self.lutsize = 1024
        self.Xcosn = (
            np.pi
            * np.array(range(-(2 * self.lutsize + 1), (self.lutsize + 2)))
            / self.lutsize
        )
        self.alpha = (self.Xcosn + np.pi) % (2 * np.pi) - np.pi

        max_ht = np.floor(np.log2(min(self.image_shape[0], self.image_shape[1]))) - 2
        if height == "auto":
            self.num_scales = int(max_ht)
        elif height > max_ht:
            raise ValueError(f"Cannot build pyramid higher than {max_ht:.0f} levels.")
        elif height < 0:
            raise ValueError("Height must be a non-negative integer.")
        else:
            self.num_scales = int(height)

        if self.order > 15 or self.order < 0:
            raise ValueError("order must be an integer in the range [0, 15].")
        if self.order == 0 and self.is_complex:
            raise ValueError(
                "Complex pyramid cannot have order=0! See "
                "https://github.com/plenoptic-org/plenoptic/issues/326 "
                "for an explanation."
            )
        self.num_orientations = int(self.order + 1)

        if twidth <= 0:
            raise ValueError("twidth must be positive.")
        twidth = int(twidth)

        dims = np.array(self.image_shape)

        # make a grid for the raised cosine interpolation
        ctr = np.ceil((np.array(dims) + 0.5) / 2).astype(int)

        (xramp, yramp) = np.meshgrid(
            np.linspace(-1, 1, dims[1] + 1)[:-1],
            np.linspace(-1, 1, dims[0] + 1)[:-1],
            indexing="xy",
        )

        self.angle = np.arctan2(yramp, xramp)
        log_rad = np.sqrt(xramp**2 + yramp**2)
        log_rad[ctr[0] - 1, ctr[1] - 1] = log_rad[ctr[0] - 1, ctr[1] - 2]
        self.log_rad = np.log2(log_rad)

        # radial transition function (a raised cosine in log-frequency):
        self.Xrcos, Yrcos = raised_cosine(twidth, (-twidth / 2.0), np.array([0, 1]))
        self.Yrcos = np.sqrt(Yrcos)

        self.YIrcos = np.sqrt(1.0 - self.Yrcos**2)

        # create low and high masks
        lo0mask = interpolate1d(self.log_rad, self.YIrcos, self.Xrcos)
        hi0mask = interpolate1d(self.log_rad, self.Yrcos, self.Xrcos)
        self.register_buffer("lo0mask", torch.as_tensor(lo0mask).unsqueeze(0))
        self.register_buffer("hi0mask", torch.as_tensor(hi0mask).unsqueeze(0))

        # need a mock image to down-sample so that we correctly
        # construct the differently-sized masks
        mock_image = np.random.rand(*self.image_shape)
        imdft = np.fft.fftshift(np.fft.fft2(mock_image))
        lodft = imdft * lo0mask

        # this list, used by coarse-to-fine optimization, gives all the
        # scales (including residuals) from coarse to fine
        self.scales = (
            ["residual_lowpass"]
            + list(range(self.num_scales))[::-1]
            + ["residual_highpass"]
        )

        # we create these copies because they will be modified in the
        # following loops
        Xrcos = self.Xrcos.copy()
        angle = self.angle.copy()
        log_rad = self.log_rad.copy()
        # pre-generate the angle, hi and lo masks, as well as the
        # indices used for down-sampling.
        self._loindices = []
        for i in range(self.num_scales):
            Xrcos -= np.log2(2)
            const = (
                (2 ** (2 * self.order))
                * (factorial(self.order, exact=True) ** 2)
                / float(self.num_orientations * factorial(2 * self.order, exact=True))
            )

            if self.is_complex:
                Ycosn_forward = (
                    2.0
                    * np.sqrt(const)
                    * (np.cos(self.Xcosn) ** self.order)
                    * (np.abs(self.alpha) < np.pi / 2.0).astype(int)
                )
                Ycosn_recon = np.sqrt(const) * (np.cos(self.Xcosn)) ** self.order

            else:
                Ycosn_forward = np.sqrt(const) * (np.cos(self.Xcosn)) ** self.order
                Ycosn_recon = Ycosn_forward

            himask = interpolate1d(log_rad, self.Yrcos, Xrcos)
            self.register_buffer(
                f"_himasks_scale_{i}", torch.as_tensor(himask).unsqueeze(0)
            )

            anglemasks = []
            anglemasks_recon = []
            for b in range(self.num_orientations):
                anglemask = interpolate1d(
                    angle,
                    Ycosn_forward,
                    self.Xcosn + np.pi * b / self.num_orientations,
                )
                anglemask_recon = interpolate1d(
                    angle,
                    Ycosn_recon,
                    self.Xcosn + np.pi * b / self.num_orientations,
                )
                anglemasks.append(torch.as_tensor(anglemask).unsqueeze(0))
                anglemasks_recon.append(torch.as_tensor(anglemask_recon).unsqueeze(0))

            self.register_buffer(f"_anglemasks_scale_{i}", torch.cat(anglemasks))
            self.register_buffer(
                f"_anglemasks_recon_scale_{i}", torch.cat(anglemasks_recon)
            )
            if not self.downsample:
                lomask = interpolate1d(log_rad, self.YIrcos, Xrcos)
                self.register_buffer(
                    f"_lomasks_scale_{i}", torch.as_tensor(lomask).unsqueeze(0)
                )
                self._loindices.append([np.array([0, 0]), dims])
                lodft = lodft * lomask

            else:
                # subsample lowpass
                dims = np.array([lodft.shape[0], lodft.shape[1]])
                ctr = np.ceil((dims + 0.5) / 2).astype(int)
                lodims = np.ceil((dims - 0.5) / 2).astype(int)
                loctr = np.ceil((lodims + 0.5) / 2).astype(int)
                lostart = ctr - loctr
                loend = lostart + lodims
                self._loindices.append([lostart, loend])

                # subsample indices
                log_rad = log_rad[lostart[0] : loend[0], lostart[1] : loend[1]]
                angle = angle[lostart[0] : loend[0], lostart[1] : loend[1]]

                lomask = interpolate1d(log_rad, self.YIrcos, Xrcos)
                self.register_buffer(
                    f"_lomasks_scale_{i}", torch.as_tensor(lomask).unsqueeze(0)
                )
                # subsampling
                lodft = lodft[lostart[0] : loend[0], lostart[1] : loend[1]]
                # convolution in spatial domain
                lodft = lodft * lomask

        # reasonable default dtype
        self.to(torch.float32)
        # This model has no trainable parameters, so it's always in eval mode
        self.eval()

    def forward(
        self,
        x: Tensor,
        scales: list[SCALES_TYPE] | None = None,
    ) -> OrderedDict:
        r"""
        Generate the steerable pyramid coefficients for an image.

        The steerable pyramid coefficients run from fine to coarse and split the image
        into subbands corresponding to different orientations and scales (i.e., spatial
        frequencies).

        Parameters
        ----------
        x
            A tensor containing the image to analyze. We want to operate
            on this in the pytorch-y way, so we want it to be 4d (batch,
            channel, height, width).
        scales
            Which scales to include in the returned representation. If None, we
            include all scales. Otherwise, can contain subset of values present
            in this model's ``scales`` attribute (ints from 0 up to
            ``self.num_scales-1`` and the strs 'residual_highpass' and
            'residual_lowpass'. Can contain a single value or multiple values.
            If it's an int, we include all orientations from that scale. Order
            within the list does not matter.

        Returns
        -------
        representation
            Pyramid coefficients. These will be stored in an ordered dictionary with
            keys that are ``"residual_highpass"``, ``"residual_lowpass"``, or tuples of
            form ``(scale, orientation)``, where ``scale`` runs from ``0`` to
            ``self.num_scales-1`` and ``orientation`` runs from 0 to ``self.order``.
            Coefficients have shape ``(*x.shape[:2], x.shape[2] / 2**scale, x.shape[3] /
            2**scale)``, with the ``"residual_highpass"`` height and width matching that
            of ``x``, and ``"residual_lowpass"`` having height and width ``(x.shape[2] /
            2**self.num_scales, x.shape[3] / 2**self.num_scales)``. They are ordered
            from fine to coarse: ``"residual_highpass", (scale=0, orientation=0),
            (scale=0, orientation=1), ..., (scale=num_scales-1, orientation=order),
            "residual_lowpass"``.

        Raises
        ------
        ValueError
            If ``image`` is the wrong shape, i.e. ``image.shape[-2:] !=
            self.image_shape``.

        Examples
        --------
        .. plot::
          :context: reset

          >>> import plenoptic as po
          >>> img = po.data.einstein()
          >>> spyr = po.simul.SteerablePyramidFreq(img.shape[-2:])
          >>> po.pyrshow(spyr(img))
          <PyrFigure ...>
        """
        if self.image_shape != x.shape[-2:]:
            raise ValueError(
                f"Input tensor height/width {tuple(x.shape[-2:])} does not match "
                f"image_shape set at initialization {tuple(self.image_shape)}. "
                "Either resize the input or re-initialize this model."
            )
        pyr_coeffs = OrderedDict()
        if scales is None:
            scales = self.scales
        scale_ints = [s for s in scales if isinstance(s, int)]
        if len(scale_ints) != 0:
            assert (max(scale_ints) < self.num_scales) and (min(scale_ints) >= 0), (
                "Scales must be within 0 and num_scales-1"
            )
        angle = self.angle.copy()
        log_rad = self.log_rad.copy()
        lo0mask = self.lo0mask.clone()
        hi0mask = self.hi0mask.clone()

        # x is a torch tensor batch of images of size (batch, channel, height,
        # width)
        assert len(x.shape) == 4, "Input must be batch of images of shape BxCxHxW"

        imdft = fft.fft2(x, dim=(-2, -1), norm=self.fft_norm)
        imdft = fft.fftshift(imdft)

        if "residual_highpass" in scales:
            # high-pass
            hi0dft = imdft * hi0mask
            hi0 = fft.ifftshift(hi0dft)
            hi0 = fft.ifft2(hi0, dim=(-2, -1), norm=self.fft_norm)
            pyr_coeffs["residual_highpass"] = hi0.real
            self.pyr_size["residual_highpass"] = tuple(hi0.real.shape[-2:])

        # input to the next scale is the low-pass filtered component
        lodft = imdft * lo0mask

        for i in range(self.num_scales):
            if i in scales:
                # high-pass mask is selected based on the current scale
                himask = getattr(self, f"_himasks_scale_{i}")
                # compute filter output at each orientation
                for b in range(self.num_orientations):
                    # band pass filtering is done in the fourier space as multiplying
                    #  by the fft of a gaussian derivative.
                    # The oriented dft is computed as a product of the fft of the
                    # low-passed component, the precomputed anglemask (specifies
                    # orientation), and the precomputed hipass mask (creating a bandpass
                    # filter) the complex_const variable comes from the Fourier
                    # transform of a gaussian derivative.
                    # Based on the order of the gaussian, this constant changes.

                    anglemask = getattr(self, f"_anglemasks_scale_{i}")[b]

                    complex_const = np.power(complex(0, -1), self.order)
                    banddft = complex_const * lodft * anglemask * himask
                    # fft output is then shifted to center frequencies
                    band = fft.ifftshift(banddft)
                    # ifft is applied to recover the filtered representation in spatial
                    # domain
                    band = fft.ifft2(band, dim=(-2, -1), norm=self.fft_norm)

                    # for real pyramid, take the real component of the complex band
                    if not self.is_complex:
                        pyr_coeffs[(i, b)] = band.real
                    else:
                        # Because the input signal is real, to maintain a tight frame
                        # if the complex pyramid is used, magnitudes need to be divided
                        # by sqrt(2) because energy is doubled.

                        if self.tight_frame:
                            band = band / np.sqrt(2)
                        pyr_coeffs[(i, b)] = band
                    self.pyr_size[(i, b)] = tuple(band.shape[-2:])

            if not self.downsample:
                # no subsampling of angle and rad
                # just use lo0mask
                lomask = getattr(self, f"_lomasks_scale_{i}")
                lodft = lodft * lomask

                # Since we don't subsample here, if we are not using
                # orthonormalization that we need to manually account for the
                # subsampling, so that energy in each band remains the same
                # the energy is cut by factor of 4 so we need to scale magnitudes
                # by factor of 2.

                if self.fft_norm != "ortho":
                    lodft = 2 * lodft
            else:
                # subsample indices
                lostart, loend = self._loindices[i]

                log_rad = log_rad[lostart[0] : loend[0], lostart[1] : loend[1]]
                angle = angle[lostart[0] : loend[0], lostart[1] : loend[1]]

                # subsampling of the dft for next scale
                lodft = lodft[:, :, lostart[0] : loend[0], lostart[1] : loend[1]]
                # low-pass filter mask is selected
                lomask = getattr(self, f"_lomasks_scale_{i}")
                # again multiply dft by subsampled mask (convolution in spatial domain)

                lodft = lodft * lomask

        if "residual_lowpass" in scales:
            # compute residual lowpass when height <=1
            lo0 = fft.ifftshift(lodft)
            lo0 = fft.ifft2(lo0, dim=(-2, -1), norm=self.fft_norm)
            pyr_coeffs["residual_lowpass"] = lo0.real
            self.pyr_size["residual_lowpass"] = tuple(lo0.real.shape[-2:])

        return pyr_coeffs

    @staticmethod
    def convert_pyr_to_tensor(
        pyr_coeffs: OrderedDict, split_complex: bool = False
    ) -> tuple[Tensor, tuple[int, bool, list[KEYS_TYPE]]]:
        r"""
        Convert coefficient dictionary to a tensor.

        The output tensor has shape (batch, channel, height, width) and is
        intended to be used in an :class:`torch.nn.Module` downstream. In the
        multichannel case, all bands for each channel will be stacked together
        (i.e. if there are 2 channels and 18 bands per channel,
        ``pyr_tensor[:,0:18,...]`` will contain the pyr responses for channel 1 and
        ``pyr_tensor[:, 18:36, ...]`` will contain the responses for channel 2). In
        the case of a complex, multichannel pyramid with ``split_complex=True``,
        the real/imaginary bands will be intereleaved so that they appear as
        pairs with neighboring indices in the channel dimension of the tensor
        (Note: the residual bands are always real so they will only ever have a
        single band even when ``split_complex=True``.)

        This only works if ``pyr_coeffs`` was created with a pyramid with
        ``downsample=False``

        Parameters
        ----------
        pyr_coeffs
            The pyramid coefficients.
        split_complex
            Indicates whether the output should split complex bands into
            real/imag channels or keep them as a single channel. This should be
            True if you intend to use a convolutional layer on top of the
            output.

        Returns
        -------
        pyr_tensor
            Tensor with shape (batch, channel, height, width). pyramid coefficients
            reshaped into tensor. The first channel will be the residual
            highpass and the last will be the residual lowpass. Each band is
            then a separate channel, going from fine to coarse (i.e., starting with all
            of scale 0's orientations, then scale 1's, etc.).
        pyr_info
            Information required to recreate the dictionary, containing the
            number of channels, if split_complex was used in this function
            call, and the list of pyramid keys for the dictionary.

        Raises
        ------
        RuntimeError
            If ``self.downsample is True``. In this case, we can't concatenate across
            scales, because each scale is a different size.

        See Also
        --------
        convert_tensor_to_pyr
            Convert tensor representation to pyramid dictionary.

        Examples
        --------
        .. plot::
          :context: reset

          >>> import plenoptic as po
          >>> img = po.data.einstein()
          >>> spyr = po.simul.SteerablePyramidFreq(img.shape[-2:], downsample=False)
          >>> coeffs = spyr(img)
          >>> coeffs_tensor, _ = spyr.convert_pyr_to_tensor(coeffs)
          >>> coeffs_tensor.shape
          torch.Size([1, 26, 256, 256])
          >>> # rearrange so that the residuals are at the end
          >>> coeffs_tensor = [
          ...     coeffs_tensor[:, 1:-1],
          ...     coeffs_tensor[:, :1],
          ...     coeffs_tensor[:, -1:],
          ... ]
          >>> po.imshow(coeffs_tensor, col_wrap=spyr.num_orientations)
          <PyrFigure ...>
        """
        pyr_keys = list(pyr_coeffs.keys())
        test_band = pyr_coeffs[pyr_keys[0]]
        num_channels = test_band.size(1)
        coeff_list = []
        for ch in range(num_channels):
            coeff_list_resid = []
            coeff_list_bands = []
            for k in pyr_keys:
                coeffs = pyr_coeffs[k][:, ch : (ch + 1), ...]
                if "residual" in k:
                    coeff_list_resid.append(coeffs)
                else:
                    if (coeffs.dtype in complex_types) and split_complex:
                        coeff_list_bands.extend([coeffs.real, coeffs.imag])
                    else:
                        coeff_list_bands.append(coeffs)

            if "residual_highpass" in pyr_coeffs:
                coeff_list_bands.insert(0, coeff_list_resid[0])
                if "residual_lowpass" in pyr_coeffs:
                    coeff_list_bands.append(coeff_list_resid[1])
            elif "residual_lowpass" in pyr_coeffs:
                coeff_list_bands.append(coeff_list_resid[0])

            coeff_list.extend(coeff_list_bands)

        try:
            pyr_tensor = torch.cat(coeff_list, dim=1)
            pyr_info = tuple([num_channels, split_complex, pyr_keys])
        except RuntimeError:
            raise RuntimeError(
                "feature maps could not be concatenated into tensor. Check that you"
                "are using coefficients that are not downsampled across scales."
                "This is done with the 'downsample=False' argument for the pyramid"
            )

        return pyr_tensor, pyr_info

    @staticmethod
    def convert_tensor_to_pyr(
        pyr_tensor: Tensor,
        num_channels: int,
        split_complex: bool,
        pyr_keys: list[KEYS_TYPE],
    ) -> OrderedDict:
        r"""
        Convert pyramid coefficient tensor to dictionary format.

        ``num_channels``, ``split_complex``, and ``pyr_keys`` are elements of
        the ``pyr_info`` tuple returned by ``convert_pyr_to_tensor``. You
        should always unpack the arguments for this function from that
        ``pyr_info`` tuple. See Examples section below.

        Parameters
        ----------
        pyr_tensor
            Shape (batch, channel, height, width). The pyramid coefficients.
        num_channels
            Number of channels in the original input tensor the pyramid was
            created for (i.e. if the input was an RGB image, this would be 3).
        split_complex
            Whether the ``pyr_tensor`` was created with complex channels split
            or not (if the pyramid was a complex pyramid).
        pyr_keys
            Tuple containing the list of keys for the original pyramid dictionary.

        Returns
        -------
        pyr_coeffs
            Pyramid coefficients in dictionary format.

        See Also
        --------
        convert_pyr_to_tensor
            Convert pyramid dictionary representation to tensor.

        Examples
        --------
        >>> import plenoptic as po
        >>> img = po.data.einstein()
        >>> spyr = po.simul.SteerablePyramidFreq(img.shape[-2:], downsample=False)
        >>> coeffs = spyr(img)
        >>> coeffs_tensor, pyr_info = spyr.convert_pyr_to_tensor(coeffs)
        >>> new_coeffs = spyr.convert_tensor_to_pyr(coeffs_tensor, *pyr_info)
        >>> all([torch.equal(v, new_coeffs[k]) for k, v in coeffs.items()])
        True
        """
        pyr_coeffs = OrderedDict()
        i = 0
        for ch in range(num_channels):
            for k in pyr_keys:
                if "residual" in k:
                    # residuals are real-valued, the complex part is all 0
                    band = pyr_tensor[:, i, ...].unsqueeze(1).real
                    i += 1
                else:
                    if split_complex:
                        band = torch.view_as_complex(
                            rearrange(
                                pyr_tensor[:, i : i + 2, ...],
                                "b c h w -> b h w c",
                            )
                            .unsqueeze(1)
                            .contiguous()
                        )
                        i += 2
                    else:
                        band = pyr_tensor[:, i, ...].unsqueeze(1)
                        i += 1
                if k not in pyr_coeffs:
                    pyr_coeffs[k] = band
                else:
                    pyr_coeffs[k] = torch.cat([pyr_coeffs[k], band], dim=1)

        return pyr_coeffs

    def _recon_levels_check(
        self, levels: Literal["all"] | list[SCALES_TYPE]
    ) -> list[SCALES_TYPE]:
        r"""
        Check whether levels arg is valid for reconstruction and return valid version.

        When reconstructing the input image (i.e., when calling :meth:`recon_pyr()`),
        the user specifies which levels to include. This makes sure those
        levels are valid and gets them in the form we expect for the rest of
        the reconstruction. If the user passes ``"all"``, this constructs the
        appropriate list (based on the values of ``pyr_coeffs``).

        Parameters
        ----------
        levels
            If ``list`` should contain some subset of integers from ``0`` to
            ``self.num_scales-1`` (inclusive) and ``"residual_highpass"`` and
            ``"residual_lowpass"`` (if appropriate for the pyramid). If ``"all"``,
            returned value will contain all valid levels.

        Returns
        -------
        levels
            List containing the valid levels for reconstruction.

        Raises
        ------
        TypeError
            If ``levels`` is not one of the allowed values.
        """
        if isinstance(levels, str):
            if levels != "all":
                raise TypeError(
                    "levels must be a list of levels or the string 'all' but"
                    f" got {levels}"
                )
            levels = (
                ["residual_highpass"]
                + list(range(self.num_scales))
                + ["residual_lowpass"]
            )
        else:
            if not hasattr(levels, "__iter__"):
                raise TypeError(
                    "levels must be a list of levels or the string 'all' but"
                    f" got {levels}"
                )
            levs_nums = np.array([int(i) for i in levels if isinstance(i, int)])
            assert (levs_nums >= 0).all(), "Level numbers must be non-negative."
            assert (levs_nums < self.num_scales).all(), (
                f"Level numbers must be in the range [0, {self.num_scales - 1:d}]"
            )
            levs_tmp = list(np.sort(levs_nums))  # we want smallest first
            if "residual_highpass" in levels:
                levs_tmp = ["residual_highpass"] + levs_tmp
            if "residual_lowpass" in levels:
                levs_tmp = levs_tmp + ["residual_lowpass"]
            levels = levs_tmp
        # not all pyramids have residual highpass / lowpass, but it's easier
        # to construct the list including them, then remove them if necessary.
        if "residual_lowpass" not in self.pyr_size and "residual_lowpass" in levels:
            levels.pop(-1)
        if "residual_highpass" not in self.pyr_size and "residual_highpass" in levels:
            levels.pop(0)
        return levels

    def _recon_bands_check(self, bands: Literal["all"] | list[int]) -> list[int]:
        """
        Check whether bands arg is valid for reconstruction and return valid version.

        When reconstructing the input image (i.e., when calling :meth:`recon_pyr()`),
        the user specifies which orientations to include. This makes sure those
        orientations are valid and gets them in the form we expect for the rest
        of the reconstruction. If the user passes ``'all'``, this
        constructs the appropriate list (based on the values of ``pyr_coeffs``).

        Parameters
        ----------
        bands
            If list, should contain some subset of integers from ``0`` to
            ``self.num_orientations-1``. If ``'all'``, returned value will
            contain all valid orientations.

        Returns
        -------
        bands
            List containing the valid orientations for reconstruction.

        Raises
        ------
        TypeError
            If ``bands`` is not an int or ``"all"``.
        ValueError
            If ``bands`` is an integer outside of the range ``[0,
            self.num_orientations-1]``.
        """
        if isinstance(bands, str):
            if bands != "all":
                raise TypeError(
                    f"bands must be a list of ints or the string 'all' but got {bands}"
                )
            bands = np.arange(self.num_orientations)
        else:
            if not hasattr(bands, "__iter__"):
                raise TypeError(
                    f"bands must be a list of ints or the string 'all' but got {bands}"
                )
            bands: NDArray = np.array(bands, ndmin=1)
            assert (bands >= 0).all(), "Error: band numbers must be larger than 0."
            if any(bands > self.num_orientations):
                raise ValueError(
                    "Error: band numbers must be in the range "
                    f"[0, {self.num_orientations - 1:d}]"
                )
        return list(bands)

    def _recon_keys(
        self,
        levels: Literal["all"] | list[SCALES_TYPE],
        bands: Literal["all"] | list[int],
    ) -> list[KEYS_TYPE]:
        """
        Make a list of all the relevant keys from to use in pyramid reconstruction.

        When reconstructing the input image (i.e., when calling :meth:`recon_pyr()`),
        the user specifies some subset of the pyramid coefficients to include
        in the reconstruction. This function takes in those specifications,
        checks that they're valid, and returns a list of tuples that are keys
        into the ``pyr_coeffs`` dictionary.

        Parameters
        ----------
        levels
            If ``list`` should contain some subset of integers from ``0`` to
            ``self.num_scales-1`` (inclusive) and ``"residual_highpass"`` and
            ``"residual_lowpass"`` (if appropriate for the pyramid). If ``"all"``,
            returned value will contain all valid levels.
        bands
            If list, should contain some subset of integers from ``0`` to
            ``self.num_orientations-1``. If ``"all"``, returned value will contain
            all valid orientations.

        Returns
        -------
        recon_keys
            List of tuples, all of which are keys in ``pyr_coeffs``. These are
            the coefficients to include in the reconstruction of the image.
        """
        levels = self._recon_levels_check(levels)
        bands = self._recon_bands_check(bands)
        recon_keys = []
        for level in levels:
            # residual highpass and lowpass
            if isinstance(level, str):
                recon_keys.append(level)
            # else we have to get each of the (specified) bands at
            # that level
            else:
                recon_keys.extend([(level, band) for band in bands])
        return recon_keys

    def recon_pyr(
        self,
        pyr_coeffs: OrderedDict,
        levels: Literal["all"] | list[SCALES_TYPE] = "all",
        bands: Literal["all"] | list[int] = "all",
    ) -> Tensor:
        """
        Reconstruct the image or batch of images, optionally using subset coefficients.

        NOTE: in order to call this function, you need to have previously called
        :meth:`forward`, with the tensor you wish to reconstruct. This will fail if
        you called ``forward()`` with a subset of scales.

        Parameters
        ----------
        pyr_coeffs
            Pyramid coefficients to reconstruct from.
        levels
            If ``list`` should contain some subset of integers from ``0`` to
            ``self.num_scales-1`` (inclusive), ``"residual_lowpass"``, and
            ``"residual_highpass"``. If ``"all"``, returned value will contain all
            valid levels. Otherwise, must be one of the valid levels.
        bands
            If list, should contain some subset of integers from ``0`` to
            ``self.num_orientations-1``. If ``"all"``, returned value will contain
            all valid orientations. Otherwise, must be one of the valid
            orientations.

        Returns
        -------
        recon
            The reconstructed image, of shape (batch, channel, height, width).

        Raises
        ------
        ValueError
            If ``self.forward()`` was called with ``scales`` argument not ``None``.
        TypeError
            If ``levels`` is not one of the allowed values.
        TypeError
            If ``bands`` is not an integer or ``"all"`` .
        ValueError
            If ``bands`` is an integer outside of the range ``[0,
            self.num_orientations-1]``.

        Examples
        --------
        .. plot::
          :context: reset

          >>> import plenoptic as po
          >>> import torch
          >>> img = po.data.einstein()
          >>> spyr = po.simul.SteerablePyramidFreq(img.shape[-2:])
          >>> coeffs = spyr(img)
          >>> recon = spyr.recon_pyr(coeffs)
          >>> torch.allclose(recon, img, rtol=1e-8, atol=1e-5)
          True
          >>> titles = ["Original", "Reconstructed", "Difference"]
          >>> po.imshow([img, recon, img - recon], title=titles)
          <PyrFigure ...>
        """
        # For reconstruction to work, last time we called forward needed
        # to include all levels
        for s in self.scales:
            if isinstance(s, str):
                if s not in pyr_coeffs:
                    raise ValueError(
                        f"scale {s} not in pyr_coeffs! pyr_coeffs must include"
                        " all scales, so make sure forward() was called with"
                        " arg scales=None"
                    )
            else:
                for b in range(self.num_orientations):
                    if (s, b) not in pyr_coeffs:
                        raise ValueError(
                            f"scale {s} not in pyr_coeffs! pyr_coeffs must"
                            " include all scales, so make sure forward() was"
                            " called with arg scales=None"
                        )

        recon_keys = self._recon_keys(levels, bands)
        scale = 0

        # load masks from model
        lo0mask = self.lo0mask
        hi0mask = self.hi0mask

        # Recursively generate the reconstruction - function starts with
        # fine scales going down to coarse and then the reconstruction
        # is built recursively from the coarse scale up

        recondft = self._recon_levels(pyr_coeffs, recon_keys, scale)

        # generate highpass residual Reconstruction
        if "residual_highpass" in recon_keys:
            hidft = fft.fft2(
                pyr_coeffs["residual_highpass"],
                dim=(-2, -1),
                norm=self.fft_norm,
            )
            hidft = fft.fftshift(hidft)

            # output dft is the sum of the recondft from the recursive
            # function times the lomask (low pass component) with the
            # highpass dft * the highpass mask
            outdft = recondft * lo0mask + hidft * hi0mask
        else:
            outdft = recondft * lo0mask

        # get output reconstruction by inverting the fft
        reconstruction = fft.ifftshift(outdft)
        reconstruction = fft.ifft2(reconstruction, dim=(-2, -1), norm=self.fft_norm)

        # get real part of reconstruction (if complex)
        reconstruction = reconstruction.real

        return reconstruction

    def _recon_levels(
        self, pyr_coeffs: OrderedDict, recon_keys: list[KEYS_TYPE], scale: int
    ) -> Tensor:
        """
        Recursive function used to build the reconstruction. Called by recon_pyr.

        Each time this function is called, it reconstructs a single scale.

        Parameters
        ----------
        pyr_coeffs
            Dictionary containing the coefficients of the pyramid. Keys are
            ``(level, band)`` tuples and the strings ``"residual_lowpass"`` and
            ``"residual_highpass"`` and values are Tensors of shape (batch,
            channel, height, width).
        recon_keys
            List of the keys that index into the pyr_coeffs Dictionary.
        scale
            Current scale that is being used to build the reconstruction
            scale is incremented by 1 on each call of the function.

        Returns
        -------
        recondft
            Current reconstruction based on the orientation band dft from the
            current scale summed with the output of recursive call with the
            next scale incremented.
        """
        # base case, return the low-pass residual
        if scale == self.num_scales:
            if "residual_lowpass" in recon_keys:
                lodft = fft.fft2(
                    pyr_coeffs["residual_lowpass"],
                    dim=(-2, -1),
                    norm=self.fft_norm,
                )
                lodft = fft.fftshift(lodft)
            else:
                lodft = fft.fft2(
                    torch.zeros_like(pyr_coeffs["residual_lowpass"]),
                    dim=(-2, -1),
                    norm=self.fft_norm,
                )

            return lodft

        # Reconstruct from orientation bands
        # update himask
        himask = getattr(self, f"_himasks_scale_{scale}")
        orientdft = torch.zeros_like(pyr_coeffs[(scale, 0)])

        for b in range(self.num_orientations):
            if (scale, b) in recon_keys:
                anglemask = getattr(self, f"_anglemasks_recon_scale_{scale}")[b]
                coeffs = pyr_coeffs[(scale, b)]
                if self.tight_frame and self.is_complex:
                    coeffs = coeffs * np.sqrt(2)

                banddft = fft.fft2(coeffs, dim=(-2, -1), norm=self.fft_norm)
                banddft = fft.fftshift(banddft)

                complex_const = np.power(complex(0, 1), self.order)
                banddft = complex_const * banddft * anglemask * himask
                orientdft = orientdft + banddft

        # get the bounding box indices for the low-pass component
        lostart, loend = self._loindices[scale]

        # create lowpass mask
        lomask = getattr(self, f"_lomasks_scale_{scale}")

        # Recursively reconstruct by going to the next scale
        reslevdft = self._recon_levels(pyr_coeffs, recon_keys, scale + 1)
        # in not downsampled case, rescale the magnitudes of the reconstructed
        # dft at each level by factor of 2 to account for the scaling in the forward
        if (not self.tight_frame) and (not self.downsample):
            reslevdft = reslevdft / 2
        # create output for reconstruction result
        resdft = torch.zeros_like(pyr_coeffs[(scale, 0)], dtype=torch.complex64)

        # place upsample and convolve lowpass component
        resdft[:, :, lostart[0] : loend[0], lostart[1] : loend[1]] = reslevdft * lomask
        recondft = resdft + orientdft
        # add orientation interpolated and added images to the lowpass image
        return recondft

    def steer_coeffs(
        self,
        pyr_coeffs: OrderedDict,
        angles: list[float],
        even_phase: bool = True,
    ) -> tuple[dict, dict]:
        """
        Steer pyramid coefficients to the specified angles.

        This allows you to have filters that have the Gaussian derivative order
        specified in construction, but arbitrary angles or number of orientations.

        Parameters
        ----------
        pyr_coeffs
            The pyramid coefficients to steer.
        angles
            List of angles (in radians) to steer the pyramid coefficients to.
        even_phase
            Specifies whether the harmonics are cosine or sine phase aligned
            about those positions.

        Returns
        -------
        resteered_coeffs
            Dictionary of re-steered pyramid coefficients. will have the same
            number of scales as the original pyramid (though it will not
            contain the residual highpass or lowpass). like ``pyr_coeffs``, keys
            are 2-tuples of ints indexing the scale and orientation, but now
            we're indexing ``angles`` instead of ``self.num_orientations``.
        resteering_weights :
            Dictionary of weights used to re-steer the pyramid coefficients.
            will have the same keys as ``resteered_coeffs``.

        Examples
        --------

        .. plot::

            >>> import plenoptic as po
            >>> import numpy as np
            >>> import torch
            >>> img = po.data.einstein()
            >>> spyr = po.simul.SteerablePyramidFreq(img.shape[-2:], height=3)
            >>> coeffs = spyr(img)
            >>> resteered_coeffs, resteering_weights = spyr.steer_coeffs(
            ...     coeffs, torch.linspace(0, 2 * np.pi, 64)
            ... )
            >>> resteered_coeffs = torch.stack(
            ...     [resteered_coeffs[(2, i + 4)] for i in range(64)], axis=2
            ... )
            >>> ani = po.animshow(resteered_coeffs, repeat=True, framerate=6, zoom=4)
            >>> # Save the video (here we're saving it as a .gif), and open it.
            >>> ani.save("resteered_coeffs.gif")

        .. image:: resteered_coeffs.gif
        """
        assert pyr_coeffs[(0, 0)].dtype not in complex_types, (
            "steering only implemented for real coefficients"
        )
        resteered_coeffs = {}
        resteering_weights = {}
        num_scales = self.num_scales
        num_orientations = self.num_orientations
        for i in range(num_scales):
            basis = torch.cat(
                [
                    pyr_coeffs[(i, j)].squeeze().unsqueeze(-1)
                    for j in range(num_orientations)
                ],
                dim=-1,
            )

            for j, a in enumerate(angles):
                res, steervect = steer(basis, a, even_phase=even_phase)
                resteering_weights[(i, j)] = steervect
                resteered_coeffs[(i, num_orientations + j)] = res.reshape(
                    pyr_coeffs[(i, 0)].shape
                )

        return resteered_coeffs, resteering_weights
